import pytest
import time

from dli.boto_authenticator import BotoAuthenticator
from unittest.mock import MagicMock


@pytest.fixture
def tmp_home(monkeypatch, tmpdir):
    # Set the 'home' dir to be a temp dir
    monkeypatch.setenv('HOME', tmpdir)
    yield tmpdir


@pytest.fixture
def client():
    client = MagicMock()
    client._environment.environment_name = 'test'
    client.session.auth_key = 'abc'
    yield client


class TestBotoAuthenticator:

    def test_modifies_existing_creds_file(self, tmp_home, client):
        tmp_home.join('.aws/credentials').ensure()
        tmp_home.join('.aws/credentials').write(
            '[default]\n'
            'aws_access_key_id = other_creds\n'
            'aws_secret_access_key = other_creds\n\n'
        )

        authenticator = BotoAuthenticator(client)
        authenticator.REFRESH_INTERVAL_SECONDS = 0.2

        authenticator.start()
        time.sleep(0.4)

        assert tmp_home.join('.aws/credentials').read() == (
            '[default]\n'
            'aws_access_key_id = other_creds\n'
            'aws_secret_access_key = other_creds\n'
            '\n'
            '[datalake_test]\n'
            'aws_access_key_id = abc\n'
            'aws_secret_access_key = noop\n\n'
        )


    def test_creates_aws_credentials_file(self, tmp_home, client):
        authenticator = BotoAuthenticator(client)
        authenticator.REFRESH_INTERVAL_SECONDS = 0.2

        authenticator.start()
        time.sleep(0.4)

        assert tmp_home.join('.aws/credentials').read() == (
            '[datalake_test]\n'
            'aws_access_key_id = abc\n'
            'aws_secret_access_key = noop\n\n'
        )

    def test_recreates_aws_credentials_file_if_deleted(self, tmp_home, client):
        authenticator = BotoAuthenticator(client)
        authenticator.REFRESH_INTERVAL_SECONDS = 0.2

        authenticator.start()
        time.sleep(0.4)
        assert tmp_home.join('.aws/credentials').exists()
        tmp_home.join('.aws/credentials').remove()
        time.sleep(0.4)
        assert tmp_home.join('.aws/credentials').exists()



import threading
import time
import sys
import os
import configparser


class BotoAuthenticator(threading.Thread):

    # As long as this is less than the margin on error between
    # the session.has_expired and the real time the credentials
    # should always be updated
    REFRESH_INTERVAL_SECONDS = 20

    def __init__(self, client):
        super().__init__()
        self.daemon = True
        self._client = client
        self._exit_flag = threading.Event()
        self.profile_name = (
            f'datalake_{self._client._environment.environment_name}'
        )
        print(
            f'Authenticator started. This will run until the DLI is closed'
            f'. This maintains up to date credentials '
            f'in your ~/.aws/credentials file under the profile '
            f'{self.profile_name}.\n'
            f'\nThis tool supports the following:\n\n'
            f'\t+--------------+-------------------------+\n'
            f'\t| Library/Tool |    Version Supported    |\n'
            f'\t+--------------+-------------------------+\n'
            f'\t| Boto3        | Any Released After 2014 |\n'
            f'\t| sdk-for-ruby | Any Released After 2014 |\n'
            f'\t| AWS-Java-SDK | 1.7.8 - 2014            |\n'
            f'\t| C++ AWS SDK  | Recent Versions         |\n'
            f'\t| Cyberduck    | Coming soon             |\n'
            f'\t| s5cmd        | All Versions            |\n'
            f'\t| Dask         | All Versions            |\n'
            f'\t| awscli       | All Versions            |\n'
            f'\t+--------------+-------------------------+\n'
        )

    def run(self):
        while True:
            try:
                self._authenticate()
                should_exit = self._exit_flag.wait(
                    timeout=self.REFRESH_INTERVAL_SECONDS
                )

                if should_exit:
                    break
            except Exception as e:
                self._client.logger.exception(
                    'Could not update credentials: ', e
                )
                break

    def stop(self):
        self._exit_flag.set()
        self._client.logger.info('Stopping Authenticator')

    def _authenticate(self):
        # This is a property lookup - so may call
        # if the session is expired.
        auth_key = self._client.session.auth_key
        shared_credentials = os.path.expanduser('~/.aws/credentials')

        # If the ~/.aws dir doesn't exists we create it
        os.makedirs(os.path.dirname(shared_credentials), exist_ok=True)

        config = configparser.ConfigParser()
        config.read(shared_credentials)

        written_access_key_id = None

        try:
            written_access_key_id = config.get(
                self.profile_name, 'aws_access_key_id'
            )
        except configparser.NoSectionError:
            pass

        # If what's on disk differs, update it
        if written_access_key_id != self._client.session.auth_key:
            self._client.logger.info(
                'Writing authentication info To ~/.aws/credentials'
            )

            config[self.profile_name] = {
                'aws_access_key_id': self._client.session.auth_key,
                'aws_secret_access_key': 'noop',
            }

            with open(shared_credentials, 'w') as configfile:
                config.write(configfile)
        else:
            self._client.logger.debug(
                'Up to date ~/.aws/credentials'
            )



#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import base64
import datetime
import os
import sys
import time
import webbrowser
from json import JSONDecodeError
import warnings
import uuid

import keyring
import jwt
import requests
import logging
import socket
from http.cookiejar import CookiePolicy

from wrapt import ObjectProxy
from http import HTTPStatus
from urllib.parse import urljoin, urlparse, parse_qs
from collections import defaultdict
from dli import __version__, __product__
from dli.analytics import AnalyticsHandler
from dli.client.adapters import DLIBearerAuthAdapter, DLIAccountsV1Adapter, \
    DLIInterfaceV1Adapter, DLISamAdapter
from dli.client.components.urls import sam_urls, identity_urls
from dli.client.environments import _Environment
from dli.client.model_descriptor import ModelDescriptor
from dli.models.organisation_model import OrganisationModel
from dli.client.components.auto_reg_metadata import AutoRegMetadata
from dli.client.components.datafile import Datafile
from dli.client.components.dataset import Dataset, Dictionary
from dli.client.components.package import Package
from dli.client.exceptions import (
    DatalakeException, InsufficientPrivilegesException,
    InvalidPayloadException, UnAuthorisedAccessException,
    CatalogueEntityNotFoundException, AuthenticationFailure,
    DatalakeGatewayTimeout, DatalakeServiceUnavailable,
    DatalakePayloadTooLargeException, TimeoutException
)
from dli.models.paginator import Paginator
from dli.models.structured_dataset_model import StructuredDatasetModel
from dli.models.unstructured_dataset_model import UnstructuredDatasetModel
from dli.modules.dataset_module import DatasetModule
from dli.modules.package_module import PackageModule
from dli.models.dictionary_model import DictionaryModel
from dli.models.file_model import FileModel
from dli.models.instance_model import InstanceModel, \
    InstancesCollection as InstancesCollection_
from dli.models.package_model import PackageModel
from dli.models.dataset_factory import DatasetFactory
from dli.siren import PatchedSirenBuilder
from dli.boto_authenticator import BotoAuthenticator

trace_logger = logging.getLogger('trace_logger')

class DliClient(AutoRegMetadata,
                Datafile, Dataset, Dictionary,
                Package):
    """
    Definition of a client. This client mixes in utility functions for
    manipulating packages, datasets and datafiles.
    """

    _DatasetFactory = ModelDescriptor(DatasetFactory)
    _Structured = ModelDescriptor(StructuredDatasetModel)
    _Unstructured = ModelDescriptor(UnstructuredDatasetModel)
    _Instance = ModelDescriptor(InstanceModel)
    _InstancesCollection = ModelDescriptor(InstancesCollection_)
    _Pagination = ModelDescriptor(Paginator)
    _File = ModelDescriptor(FileModel)
    _Package = ModelDescriptor(PackageModel)
    _Organisation = ModelDescriptor(OrganisationModel)
    _DictionaryV2 = ModelDescriptor(DictionaryModel)

    _packages = ModelDescriptor(PackageModule)
    _datasets = ModelDescriptor(DatasetModule)

    _environment_class = _Environment

    def __init__(self, api_root, host=None, debug=None, strict=True,
                 access_id=None, secret_key=None, use_keyring=True):
        self._environment = self._environment_class(api_root)
        self.access_id = access_id
        self.secret_key = secret_key
        self.host = host
        self.debug = debug
        self.strict = strict
        self.use_keyring = use_keyring
        self.logger = logging.getLogger(__name__)

        self.logger.debug(
            'Starting SDK session',
            extra={
               'catalogue': self._environment.catalogue,
               'consumption': self._environment.consumption,
               'strict' : strict,
               'version': __version__
            }
        )

        self._session = self._new_session()
        self._analytics_handler = AnalyticsHandler(self)

        self.packages = self._packages()
        self.datasets = self._datasets()

        if access_id is None and secret_key is not None:
            warnings.warn(
                'The parameter `api_key` will be deprecated in the future. '
                'We will be contacting users in the following months to '
                'explain how to migrate to the new authentication flow.',
                PendingDeprecationWarning
            )

        self._boto_authenticator = None


    def _new_session(self):
        session = Session(
            self.access_id,
            self.secret_key,
            self._environment,
            self.host,
            logger=self.logger,
            use_keyring=self.use_keyring
        )
        return session

    @property
    def session(self):
        # if the session expired, then reauth
        # and create a new context
        if self._session.has_expired:
            self._session = self._new_session()
        return self._session

    @property
    def aws_authenticator(self):
        """
        Worker object that can be called with:

        >>> client.aws_authenticator.start()

        In order to sync the ~/.aws/credentials file with
        the current user. Can be explicitly stopped via:

        >>> client.aws_authenticator.stop()

        Will simply shutdown when the DLI exits.
        """
        if (
            not self._boto_authenticator or
            not self._boto_authenticator.is_alive()
        ):
            self._boto_authenticator = BotoAuthenticator(self)

        return self._boto_authenticator


class Session(requests.Session):

    # Full list of possible algorithms:
    # https://pyjwt.readthedocs.io/en/latest/algorithms.html
    __pyjwt_algorithms = [
        # `HS256` is the value returned by the JWT from prod, but pyjwt seems
        # to complain about the signature.
        'HS256',
        'HS512', 'ES256', 'ES384', 'ES512', 'RS256', 'RS384',
        'RS512', 'PS256', 'PS384', 'PS512',
    ]

    def __init__(
        self, access_id, secret_key, environment, host, logger,
        auth_key=None, auth_prompt=True, use_keyring=True
    ):
        super().__init__()
        self.auth_key = None
        self.logger = logger
        self.access_id = access_id
        self.secret_key = secret_key
        self._environment = environment
        self.use_keyring = use_keyring
        self.host = host
        self.siren_builder = PatchedSirenBuilder()
        if auth_prompt:
            self._auth_init(auth_key)
        # mount points to add headers to specific routing requests
        self._set_mount_adapters()

        # Don't allow cookies to be set.
        # The new API will reject requests with both a cookie
        # and a auth header (as there's no predictable crediential to choose).
        #
        # However the old API, once authenticated using a Bearer token, will
        # as a side effect of a request return a oidc_id_token which matches
        # the auth header. This is ignored.
        self.cookies.set_policy(BlockAll())

    def __check(self, issplit, splitlen):
        check_key = keyring.get_password(
            __product__, self._environment.catalogue)
        if check_key:
            split_check = check_key.split("**split**")
            if len(split_check) > 1:
                issplit = True
                splitlen = int(split_check[-1])
                chunked_jwt = ''
                # we must re-form the jwt
                for ix in range(splitlen):
                    chunked_jwt += keyring.get_password(
                        __product__,
                        self._environment.catalogue + f"-{ix}")

                check_key = chunked_jwt
        return check_key, issplit, splitlen

    def __exception_handler(self, nke, raises):
        # disable the use of keyring
        self.use_keyring = False
        print(
            "We have noticed you are trying to run the SDK on a headless "
            "machine but your application credential environment variables "
            "(DLI_ACCESS_KEY_ID and DLI_SECRET_ACCESS_KEY) "
            "are not set or there is no keyring manager available. \n\n"
            f"YOU MAY NEED TO LOGIN EVERY TIME. \n"
            f"Please contact support at datalake-support@ihsmarkit.com"
            f"if you need more help. \n\n"
            f"{str(nke)}\n\n"
        )
        if raises:
            raise nke

    def _reload_or_web_flow(self, raises=False):
        from keyring.errors import PasswordDeleteError

        check_key = None
        issplit = False
        splitlen = 0

        if hasattr(self, "use_keyring") and self.use_keyring:
            try:
                from keyring.errors import NoKeyringError
                try:
                    check_key, issplit, splitlen = self.__check(issplit, splitlen)
                except NoKeyringError as nke:
                    self.__exception_handler(nke, raises)
            except ImportError:
                try:
                    check_key, issplit, splitlen = self.__check(issplit, splitlen)
                except RuntimeError as nke:
                    self.__exception_handler(nke, raises)

        reloaded = None
        if check_key:
            try:
                reloaded = check_key
                decoded = jwt.decode(
                    jwt=str.encode(reloaded),
                    # `verify` is deprecated in pyjwt 1.5.1, use `options`
                    # param https://pyjwt.readthedocs.io/en/latest/api.html?
                    # highlight=verify_signature#jwt.decode
                    verify=False,
                    algorithms=self.__pyjwt_algorithms,
                    options={
                        "verify_signature": False,
                    }
                )
                if decoded.get("exp", 0) <= time.time():
                    try:
                        keyring.delete_password(__product__,
                                                self._environment.catalogue)
                        if issplit:
                            for ix in range(splitlen):
                                keyring.delete_password(
                                    __product__,
                                    self._environment.catalogue + f"-{ix}"
                                )
                    except PasswordDeleteError as e:
                        self.logger.debug("No such password")
                    finally:
                        reloaded = None
            except Exception as e:
                raise e

        if reloaded:
            self.logger.debug(f"Using vault JWT "
                             f"{__product__} "
                             f"{self._environment.catalogue}")
            self.auth_key = reloaded
        elif not self.secret_key and not self.access_id:
            self.auth_key = self._get_web_auth_key()

    def _auth_init(self, auth_key=None):

        if auth_key:
            self.auth_key = auth_key
        elif self.secret_key and not self.access_id:
            self.auth_key = self._get_auth_key()
        elif self.access_id and self.secret_key:
            self.auth_key = self._get_SAM_auth_key()
        else:
            self._reload_or_web_flow()

        self.decoded_token = self._get_decoded_token()
        self.token_expires_on = self._get_expiration_date()

    def _get_timeout(self, timeout=None):
        # default timeout is 60.
        return timeout

    def request(self, method, url, *args, **kwargs):
        if not urlparse(url).netloc:
            url = urljoin(self._environment.catalogue, url)

        kwargs.pop('hooks', None)
        hooks = {'response': self._response_hook}
        timeout = self._get_timeout()

        headers = kwargs.pop('headers', {})
        # Adding this here in order to log the request ID
        request_id = str(uuid.uuid4())
        headers['X-Request-ID'] = request_id

        try:
            self.logger.debug(
                f'{method} Request to {url} with request_id: {request_id}',
                extra={
                    'method': method,
                    'request': url,
                    '-args': args,
                    '-kwargs': kwargs,
                    'request_id': request_id
                }
            )

            return super().request(
                method, url, hooks=hooks, timeout=timeout, headers=headers, *args, **kwargs
            )
        except requests.exceptions.ConnectionError as e:
            raise TimeoutException(e)
        except socket.error as e:
            raise ValueError(
                'Unable to process request due to a networking issue '
                'root cause could be a bad connection, '
                'not being on the correct VPN, '
                'or a network timeout '
            ) from e

    @property
    def has_expired(self):
        # We subtract timedelta from the expiration time in order to allow a safety window for
        # a code block to execute after a check has been asserted.
        return datetime.datetime.utcnow() > \
               (self.token_expires_on - datetime.timedelta(minutes=1))

    def _response_hook(self, response, *args, **kwargs):
        # Apologies for the ugly code. The startswith siren check
        # is to make this only relevant to the old API.
        response = Response(response, self.siren_builder)


        if self.logger:
            self.logger.debug(
                'Response',
                extra={
                    # 'content': response.content,
                    'status': response.status_code,
                    'method': response.request.method,
                    'request': response.request.url,
                    'headers': response.request.headers
                }
            )

        if not response.ok:
            exceptions = defaultdict(
                lambda: DatalakeException,
                {
                 # 400 errors
                 HTTPStatus.BAD_REQUEST: InvalidPayloadException,
                 HTTPStatus.UNPROCESSABLE_ENTITY: InvalidPayloadException,
                 HTTPStatus.UNAUTHORIZED: UnAuthorisedAccessException,
                 HTTPStatus.FORBIDDEN: InsufficientPrivilegesException,
                 HTTPStatus.NOT_FOUND: CatalogueEntityNotFoundException,

                 # 500 errors
                 HTTPStatus.REQUEST_ENTITY_TOO_LARGE: DatalakePayloadTooLargeException,
                 HTTPStatus.SERVICE_UNAVAILABLE: DatalakeServiceUnavailable,
                 HTTPStatus.BAD_GATEWAY: DatalakeServiceUnavailable,
                 HTTPStatus.GATEWAY_TIMEOUT: DatalakeGatewayTimeout,
                 }
            )


            try:
                message = response.json()
            except (JSONDecodeError, ValueError):
                message = response.text

            raise exceptions[response.status_code](
                message=message,
                params=parse_qs(urlparse(response.request.url).query),
                response=response
            )

        return response

    def _set_mount_adapters(self):
        self.mount(
            urljoin(self._environment.catalogue, '__api/'),
            DLIInterfaceV1Adapter(self)
        )

        self.mount(
            urljoin(self._environment.catalogue, '__api_v2/'),
            DLIBearerAuthAdapter(self)
        )

        self.mount(
            self._environment.consumption, DLIBearerAuthAdapter(self)
        )

        self.mount(
            urljoin(self._environment.accounts, 'api/identity/v1/'),
            DLIAccountsV1Adapter(self)
        )

        self.mount(
            urljoin(self._environment.accounts, 'api/identity/v2/'),
            DLIBearerAuthAdapter(self)
        )

        self.mount(
            self._environment.sam, DLISamAdapter(self)
        )

        self.mount(
            self._environment.consumption, DLIBearerAuthAdapter(self)
        )

    def _get_decoded_token(self):
        return jwt.decode(
            jwt=self.auth_key,
            # `verify` is deprecated in pyjwt 1.5.1, use `options`
            # param https://pyjwt.readthedocs.io/en/latest/api.html
            # ?highlight=verify_signature#jwt.decode
            verify=False,
            algorithms=self.__pyjwt_algorithms,
            options={
                "verify_signature": False,
            }
        )

    def _get_expiration_date(self):
        default_timeout = (
            datetime.datetime.utcnow() +
            datetime.timedelta(minutes=55)
        )

        if 'exp' not in self.decoded_token:
            return default_timeout

        return datetime.datetime.utcfromtimestamp(
            self.decoded_token['exp']
        ) - datetime.timedelta(minutes=5)

    def _get_auth_key(self):
        # todo - calling this when JWT from web flow fails.
        try:
            response = self.post(
                '/__api/start-session',
                headers={
                    'Authorization': 'Bearer {}'.format(
                        self.secret_key
                    )
                }
            )
        except DatalakeException as e:
            raise AuthenticationFailure(
                message='Could not authenticate API key'
            ) from e

        return response.text

    def _get_SAM_auth_key(self):

        sam_response = self.post(
            urljoin(self._environment.sam, sam_urls.sam_token),
            data={
                "client_id": self.access_id,
                "client_secret":  self.secret_key,
                "grant_type": "client_credentials"
            },
            hooks={'response': self._response_hook}
        )

        token = sam_response.json()["access_token"]

        catalogue_response = self.post(
            urljoin(self._environment.accounts,
                    identity_urls.identity_token),
            data={
                "client_id": self.access_id,
                "subject_token": token,
                "origin": 'SDK'
            },
            hooks={'response': self._response_hook},
        )

        _jwt = catalogue_response.json()["access_token"]
        return _jwt

    def _get_web_auth_key(self, callback=None):

        postbox = base64.urlsafe_b64encode(uuid.uuid4().bytes)\
            .decode('utf-8').replace('=', '')
        str_jwt = None

        if callback is None:
            url = f"{self._environment.catalogue}" \
                  f"{identity_urls.identity_postbox}" \
                  f"?postbox={postbox}&" \
                  f"target=/api/identity/v2/auth/postbox_confirmation"
            # we need to check that the listener can actually listen
            # before we open the browser - so establish that a listener
            # is up and running, or can listen (not firewalled)
            # else we need to tell the user that their firewall is shut.

            isopen=webbrowser.open(url, new=1)
            if not isopen:
                print(f"WARNING: We could not open a usable web browser for authentication purposes, or alternatively,"
                      f"Please copy the below:\n\n{url}\n\nEnter this into a web browser that you can access, "
                      f"or alternatively, you can use your phone or a web browser on another machine. "
                      f"After this step, by returning here, you will see this console un-blocked.\n"
                      f"If you wish to abort enter Ctrl/Cmd + C on your keyboard.\n\n")
        else:
            callback(0, postbox)

        # Cloudfront caches the calls to this URL with a TTL of
        # one minute. The result is that if the first response is a
        # HTTP status 404 then it keeps that result in cache for one
        # minute. Setting the headers to disable cache just didn't make
        # a difference, but adding a pause before the first call did.
        time.sleep(1)

        while True:
            try:
                query = self.get(
                    f"{self._environment.catalogue}"
                    f"{identity_urls.identity_poll}"
                    f"?postbox={postbox}",
                )

                if query.ok:
                    str_jwt = query.text
                    _jwt = str.encode(str_jwt)
                    break
            except TimeoutException as e:
                print(
                    "Sorry, the connection timed out. Please try again in "
                    "one minute. If this is your first time trying to connect "
                    "and the problem continues then it is possible you have "
                    "a firewall issue (please contact your IT team to "
                    "check if a firewall rule needs to be added, see "
                    "https://supportlibrary.ihsmarkit.com/display/DLSL/"
                    "Firewalls+and+proxies ).\n\n"
                    "Technical details:"
                    f"{e.message}"
                )
                sys.exit(0)

            except DatalakeException as e:
                # anti-pattern time - to be able to patch session appropriately
                # i need to modify the get, but the responsehook sees the 404
                # prior to the above logic
                if e.response.status_code == 404 or e.response.status_code == 409:
                    time.sleep(0.5)
                elif e.response.status_code == 401:
                    print("We couldn't log you in. A problem was encountered\n"
                          "If this issue persists please contact DL-Support")
                    sys.exit(0)

        if str_jwt and hasattr(self, "use_keyring") and self.use_keyring:
            try:
                keyring.set_password(__product__, self._environment.catalogue,
                                     str_jwt)
            except (Exception, OSError) as e:
                # length issue on Windows > 1280 chars in Windows Credentials
                if os.name == 'nt':
                    lens = 1000
                    chunks = [
                        str_jwt[i:i + lens]
                        for i in range(0, len(str_jwt), lens)
                    ]
                    try:
                        # index
                        keyring.set_password(
                            __product__,
                            self._environment.catalogue,
                            f"**split**{len(chunks)}"
                        )

                        # segments
                        for ix, chunk in enumerate(chunks):
                            keyring.set_password(
                                __product__,
                                self._environment.catalogue + f"-{ix}",
                                chunk
                            )
                    except Exception as e:
                        logging.warning(str(e))
                else:
                    logging.warning(str(e))
        return str_jwt


class Response(ObjectProxy):

    def __init__(self, wrapped, builder, *args, **kwargs):
        super(Response, self).__init__(wrapped, *args, **kwargs)
        self.builder = builder


class BlockAll(CookiePolicy):
    return_ok = set_ok = domain_return_ok = path_return_ok = lambda self, *args, **kwargs: False
    netscape = True
    rfc2965 = hide_cookie2 = False



#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
from deprecated import deprecated
from dli.client.components.urls import dataset_urls
from dli.client.model_descriptor import ModelDescriptor
from dli.models.structured_dataset_model import StructuredDatasetModel
from dli.models.unstructured_dataset_model import UnstructuredDatasetModel


class DatasetFactory:

    @classmethod
    @deprecated
    def _from_v1_response_to_v2(cls, v1_response):
        response = cls._client.session.get(
            dataset_urls.v2_by_id.format(
                id=v1_response['properties']['datasetId']
            )
        )

        return cls._from_v2_response(response.json())

    @classmethod
    def _from_v2_response(cls, response_json, warn=False, page_size=25):
        return cls._construct_dataset_using(
            response_json['data']['attributes'], response_json['data']['id'], warn, page_size=page_size,
        )

    @classmethod
    def _from_v2_response_unsheathed(cls, response_json, warn=False, page_size=25):
        return cls._construct_dataset_using(
            response_json['attributes'], response_json['id'], warn, page_size=page_size,
        )

    @classmethod
    def _from_v2_list_response(cls, response_json, page_size=25):
        return [
            cls._construct_dataset_using(
                dataset['attributes'], dataset['id'], page_size=page_size,
            )
            for dataset in response_json['data']
        ]

    @classmethod
    def _construct_dataset_using(cls, attributes, dataset_id, warn=False, page_size=25):
        location = attributes.pop('location')
        # In the interests of not breaking backwards compatability.
        # TODO find a way to migrate this to the new nested API.
        if not location:
            location = None
        else:
            location = location[next(iter(location))]

        # ca depend
        if attributes["content_type"] == "Unstructured":
            return cls._client._Unstructured(
                **attributes,
                location=location,
                dataset_id=dataset_id,
                warn=warn
            )
        else:
            return cls._client._Structured(
                **attributes,
                page_size=page_size,
                location=location,
                dataset_id=dataset_id
            )



#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import datetime
import http.client
import json
import logging
import textwrap
import uuid
import warnings
from collections import defaultdict
from functools import partial
from typing import List, Optional, Tuple, Union, Dict, Set
from urllib.parse import urljoin

import pandas
import pyarrow

from dli.aws import match_partitions, get_most_recent_common_prefix, \
    _get_partitions_in_filepath
from dli.client.aspects import analytics_decorator, logging_decorator
from dli.client.components.urls import consumption_urls, dataset_urls
from dli.client.exceptions import DataframeStreamingException
from dli.models import log_public_functions_calls_using, SampleData
from dli.models.dictionary_model import DictionaryModel
from dli.models.dataset_model import DatasetModel
from pyarrow.lib import ArrowInvalid
from tabulate import tabulate
from urllib3.exceptions import ProtocolError  # noqa: I900

trace_logger = logging.getLogger('trace_logger')


class StructuredDatasetModel(DatasetModel):

    def __init__(self, page_size=25, **kwargs):
        super().__init__(**kwargs,)
        self.instances = self._client._InstancesCollection(dataset=self, page_size=page_size)
        self.fields_metadata = None

    def __repr__(self):
        return f'<Structured Dataset short_code={self.short_code}>'

    def __str__(self):
        separator = "-" * 80
        splits = "\n".join(textwrap.wrap(self.description))

        return f"\nDATASET \"{self.short_code}\" [{self.data_format}]\n" \
               f">> Shortcode: {self.short_code}\n"\
               f">> Available Date Range: {self.first_datafile_at} to {self.last_datafile_at}\n" \
               f">> ID: {self.id}\n" \
               f">> Published: {self.publishing_frequency} by {self.organisation_name}\n" \
               f">> Accessible: {self.has_access}\n" \
               f"\n" \
               f"{splits}\n" \
               f"{separator}"

    def __custom_filter(
        self,
        partitions: List[str],
        paths: List[Tuple[str, int]],
    ) -> List[Tuple[str, int]]:
        """
        Filters that apply to Structured datasets
        """

        trace_logger.info(
            f"Number of paths on S3 for this dataset "
            f"after filtering for load type: '{len(paths)}'"
        )

        if partitions:
            # Only return paths that match the partition(s) provided.
            paths = [
                path for path in paths if match_partitions(path[0], partitions)
            ]

        return paths

    @property
    def sample_data(self) -> SampleData:
        return SampleData(self)

    def dask_dataframe(
        self,
        # Parameters for reading from .parquet and .csv format files:
        partitions: Optional[Union[str, List[str]]] = None,
        # Parameters for reading from .parquet format files:
        columns=None,
        filters=None,
        categories=None,
        index=None,
        engine="auto",
        gather_statistics=None,
        split_row_groups=None,
        chunksize=None,
        # Parameters for reading from .csv format files:
        blocksize='default',
        lineterminator=None,
        compression=None,
        sample=256000,
        enforce=False,
        assume_missing=False,
        include_path_column=False,
        # Parameters for reading from .parquet and .csv formats:
        **kwargs
    ):
        """
        Read a dataset into a Dask DataFrame.

        WARNING: The dependency Pandas 1.0.4 has an issue reading parquet due
        to an pandas library mismatch. This was fixed in Pandas 1.0.5. We
        advise users to not use Pandas 1.0.4.

        Background on Dask:

        A Dask DataFrame can be used to compute a Pandas Dataframe. Dask has
        the advantage of being able to process data that would be too large
        to fit into memory in a single Pandas Dataframe. The use case is to
        read the files with Dask, apply operations such as filtering and
        sorting to the Dask Dataframe (which implements most of the Pandas
        API) and finally compute the result to a Pandas Dataframe. As you
        have already done the filtering in Dask, the resulting Pandas
        Dataframe will be smaller in memory than if you tried to do
        everything in a single Pandas Dataframe.

        * Memory:

            *   Like Spark, when low on RAM it can save intermediate results to a temp directory on disk. This means more data can be processed.
            *   Copies of data will be removed from RAM when they are written to disk. Dask has the advantage of being able to process data that would be too large to fit into memory in a single Pandas Dataframe. As you have done the filtering in Dask, the Pandas Dataframe will be smaller in memory than if you tried to do everything in a single Pandas Dataframe.
            *   Lazy evaluation - Data is processed only when it is needed (like Spark). This means that less data needs to be in memory at one time before processing begins, unlike Pandas.

        * Speed - Can use a local or a remote cluster to process data in parallel. A cluster of 3 nodes would in an ideal world be 3 times faster (but usually it will be less due to network transfer speed between nodes). A Jupyter Notebook plugin is available to visualise the cluster. It is possible to have the nodes on your:

            * Local laptop.
            * Remote laptops (so a group of laptops can share the load).
            * Kubernetes cluster.

        *	Schema mismatch - clear instructions are given to the user for solving schema type mismatch. However, this is not automatically resolved (unlike Spark) and requires the user to add a `dtype` parameter (explained in the example below). Unfortunatly out-of-the-box it uses the pyarrow engine (the same as Pandas) which does not automatically support additive schemas. For dealing with additive schemas, you can:

            1. Change your code to read one file at a time as a Dask DataFrame, then `concat` the DataFrames together. The `concat` behaviour is the same as Pandas and will merge the schemas.
            2. Or switch the `engine` parameter from `pyarrow` to `fastparquet` and install `pip install fastparquet python-snappy`. Note this is difficult to install on Windows 10.

        *	API stability - Supports the API for Pandas, however they promise more stability between versions so if Pandas change something in their API, the Dask API should not change.
        *	Community support - Enterprise support is available through Coiled.io which is run by the creators of Dask.

        You can use the SDK to get a dataset, and then use the `dask_dataframe` endpoint which will return a Dask task that is setup to read the data files directly from S3 without downloading the files to the local machine.

        Please install or upgrade the SDK using the command `pip install dli[dask]` so that all of the dask plugins are installed.

        This endpoint can read data files in the following file formats:

            * .parquet
            * .csv

        into a Dask.dataframe, one file per partition.  It selects the index among the sorted columns if any exist.

        Examples
        --------

        .. code-block:: python

            dataset.dask_dataframe()

        The Dask dataframe will calculate the tasks that need to be run, but
        it does not evaluate until you call an action on it. To run a full
        evaluation across the whole dataset and return a Pandas dataframe, run:

        .. code-block:: python

            dataset.dask_dataframe().compute()

        To run the evaluation and only return the first ten results:

        .. code-block:: python

            dataset.dask_dataframe().compute(10)

        An additional parameter needs to be added when you are reading from
        compressed .csv files, for example the .csv.gz format.

        .. code-block:: python

            dataset.dask_dataframe(compression='gzip')

        Some datasets have problems with the column types changing between
        different data files. Dask will stop with an error message that
        explains the types you should pass. For example, if the error message
        that looks like this:

        /*
        ValueError: Mismatched dtypes found in `pd.read_csv`/`pd.read_table`.

        +------------+---------+----------+
        | Column     | Found   | Expected |
        +------------+---------+----------+
        | STATE_FIPS | float64 | int64    |
        +------------+---------+----------+

        Usually this is due to dask's dtype inference failing, and
        *may* be fixed by specifying dtypes manually by adding:

        dtype={'STATE_FIPS': 'float64'}
        */

        then you should do as the message says and add the parameters, then
        re-run. In this case the fix would look like this:

        .. code-block:: python

            dataset.dask_dataframe(dtype = {'STATE_FIPS': 'float64'})


        You are able to provide extra parameters to this endpoint that will
        be used by dask.

        The following parameters are used to reading from
        .parquet and .csv format files:

        Parameters
        ----------
        :param List[str] partitions: Optional. A list of filters (partitions)
            to apply to the file paths on S3.

            This allows the user to filter the files they want to read from
            before they  have to read any data. This will reduce the amount
            of redundant data that is transferred over the network and
            reduce the amount of memory required on the user's machine.

            For any dataset, the paths on S3 may contain
            partitions that are key-value pairs in the format
            `<key>=<value>`, for example in the path:

                s3://example/as_of_date=2020-01-01/country=US/file.extension

            there are the partitions:

            * `country=US`
            * `as_of_date=2020-01-01`.

            You can find the available list of partitions for a dataset by
            calling the dataset's `.partitions()` endpoint.

            You can specify a list of partitions you want to get data from by
            passing in a list in the format `<key><operator><value>`. Please
            note that you should not include any extra quotation marks around
            the value, so `as_of_date=2020-01-01` is valid but
            `as_of_date='2020-01-01'` is invalid.

            The supported operators are: `<=`, `<`, `=`, `>`, `>=`.

            For a partition where the key is named `as_of_date`, the value
            is treated as a Python datetime object, so comparisons are done
            between dates. This means that the `>=` and `<=` operators can
            be used to get data in a date range where more than one
            `as_of_date` exists. For example, to get data from only dates in
            January 2020:

            .. code-block:: python

                dataset.dask_dataframe(partitions=[
                    'as_of_date>=2020-01-01',
                    'as_of_date<=2020-01-31',
                ])

            Other key names will have their values treated as string, so the
            operators will perform Python string comparisons.

            .. code-block:: python

                dataset.dask_dataframe(partitions=[
                    'country=US'
                ])

            A list of partitions is applied as a logical AND. So specifying
            `as_of_date` and `country` will match only those paths that
            match both `as_of_date` AND `country`.

            .. code-block:: python

                dataset.dask_dataframe(partitions=[
                    'as_of_date>=2020-01-01',
                    'country=US',
                ])


        The following parameters are used to reading from
        .parquet format files:

        Parameters
        ----------
        columns : string, list or None (default)
            Field name(s) to read in as columns in the output. By default all
            non-index fields will be read (as determined by the pandas parquet
            metadata, if present). Provide a single field name instead of a list to
            read in the data as a Series.
        filters : Union[List[Tuple[str, str, Any]], List[List[Tuple[str, str, Any]]]]
            List of filters to apply, like ``[[('x', '=', 0), ...], ...]``. This
            implements partition-level (hive) filtering only, i.e., to prevent the
            loading of some row-groups and/or files.

            Predicates can be expressed in disjunctive normal form (DNF). This means
            that the innermost tuple describes a single column predicate. These
            inner predicates are combined with an AND conjunction into a larger
            predicate. The outer-most list then combines all of the combined
            filters with an OR disjunction.

            Predicates can also be expressed as a List[Tuple]. These are evaluated
            as an AND conjunction. To express OR in predictates, one must use the
            (preferred) List[List[Tuple]] notation.
        index : string, list, False or None (default)
            Field name(s) to use as the output frame index. By default will be
            inferred from the pandas parquet file metadata (if present). Use False
            to read all fields as columns.
        categories : list, dict or None
            For any fields listed here, if the parquet encoding is Dictionary,
            the column will be created with dtype category. Use only if it is
            guaranteed that the column is encoded as dictionary in all row-groups.
            If a list, assumes up to 2**16-1 labels; if a dict, specify the number
            of labels expected; if None, will load categories automatically for
            data written by dask/fastparquet, not otherwise.
        engine : {'auto', 'fastparquet', 'pyarrow'}, default 'auto'
            Parquet reader library to use. If only one library is installed, it
            will use that one; if both, it will use 'fastparquet'
        gather_statistics : bool or None (default).
            Gather the statistics for each dataset partition. By default,
            this will only be done if the _metadata file is available. Otherwise,
            statistics will only be gathered if True, because the footer of
            every file will be parsed (which is very slow on some systems).
        split_row_groups : bool or int
            Default is True if a _metadata file is available or if
            the dataset is composed of a single file (otherwise defult is False).
            If True, then each output dataframe partition will correspond to a single
            parquet-file row-group. If False, each partition will correspond to a
            complete file.  If a positive integer value is given, each dataframe
            partition will correspond to that number of parquet row-groups (or fewer).
            Only the "pyarrow" engine supports this argument.
        chunksize : int, str
            The target task partition size.  If set, consecutive row-groups
            from the same file will be aggregated into the same output
            partition until the aggregate size reaches this value.
        **kwargs: dict (of dicts)
            Passthrough key-word arguments for read backend.
            The top-level keys correspond to the appropriate operation type, and
            the second level corresponds to the kwargs that will be passed on to
            the underlying `pyarrow` or `fastparquet` function.
            Supported top-level keys: 'dataset' (for opening a `pyarrow` dataset),
            'file' (for opening a `fastparquet` `ParquetFile`), 'read' (for the
            backend read function), 'arrow_to_pandas' (for controlling the arguments
            passed to convert from a `pyarrow.Table.to_pandas()`)

        The following parameters are used to reading from
        .csv format files:

        Parameters
        ----------

        blocksize:str, int or None, optional
            Number of bytes by which to cut up larger files. Default value is
            computed based on available physical memory and the number of
            cores, up to a maximum of 64MB. Can be a number like 64000000` or
            a string like ``"64MB". If None, a single block is used for each
            file.
        lineterminator:str (length 1), optional
            Character to break file into lines. Only valid with C parser.
        compression: {‘infer’, ‘gzip’, ‘bz2’, ‘zip’, ‘xz’, None}, default
            ‘infer’
            For on-the-fly decompression of on-disk data.
        sample:int, optional
            Number of bytes to use when determining dtypes
        enforce:bool
        assume_missing:bool, optional
            If True, all integer columns that aren’t specified in dtype are
            assumed to contain missing values, and are converted to floats.
            Default is False.
        include_path_column:bool or str, optional
            Whether or not to include the path to each particular file. If
            True a new column is added to the dataframe called path. If str,
            sets new column name. Default is False.
        **kwargs
            Extra keyword arguments. See the docstring for pandas.read_csv()
            for more information on available keyword arguments.
        """
        try:
            import dask.dataframe as dd
        except ImportError:
            raise RuntimeError(
                'Before you can use Dask it must be installed into '
                'your virtual environment.'
                '\nNote: If you are running you code in a Jupyter Notebook, '
                'then before installing you will need to quit the '
                '`jupyter notebook` process that is running on the command '
                'line and start it again so that your notebooks see the '
                'version of Dask you have now installed.'
                '\nPlease install dask using one of the following options:'
                '\n1. Run `pip install dli[dask]` to install this SDK '
                'with a compatible version of dask.'
                '\n2. Run `pip install dask[dataframe]` to install the latest '
                'version of dask (which may be untested with this version '
                'of the SDK).'
            )

        self._check_access()

        start = datetime.datetime.now()
        request_id = str(uuid.uuid4())

        paths = self.__list(
            request_id=request_id,
            partitions=partitions
        )

        storage_options = {
            'client_kwargs': {
                'endpoint_url': self._DatasetModel__endpoint_url(),
            },
            'key': self._client.session.auth_key,
            'secret': 'noop',
        }

        data_format = self.data_format.lower()

        trace_logger.info(
            'dask_dataframe endpoint:'
            f"\nrequest_id: '{request_id}'"
            f"\nendpoint_url: '{self._DatasetModel__endpoint_url()}'"
            f"\ndata_format: '{data_format}'"
            '\nlisting paths elapsed time: '
            f"'{datetime.datetime.now() - start}'"
        )

        if data_format == 'parquet':
            return dd.read_parquet(
                path=paths,
                columns=columns,
                filters=filters,
                categories=categories,
                index=index,
                storage_options=storage_options,
                engine=engine,
                gather_statistics=gather_statistics,
                split_row_groups=split_row_groups,
                chunksize=chunksize,
                **kwargs
            )
        elif data_format == 'csv':
            return dd.read_csv(
                urlpath=paths,
                blocksize=blocksize,
                lineterminator=lineterminator,
                compression=compression,
                sample=sample,
                enforce=enforce,
                assume_missing=assume_missing,
                storage_options=storage_options,
                include_path_column=include_path_column,
                **kwargs
            )
        else:
            print(
                f"Sorry, the dataset is in the format '{data_format}'. This "
                'endpoint is only setup to handle parquet and csv.'
            )
            return None

    def partitions(self) -> Dict[str, List[str]]:
        """
        Retrieves the list of available partitions for a given dataset.

        The data onboarding team have structured the file paths on S3 with
        simple partitions e.g. `as_of_date` or `location`.

        Their aim was to separate the data to reduce the size of the
        individual files. For example, data that has a `location` column with
        the options `us`, `eu` and `asia` can be separated into S3 paths like
        so:

        .. code-block::

            package-name/dataset/as_of_date=2019-09-10/location=eu/filename.csv
            package-name/dataset/as_of_date=2019-09-10/location=us/filename.csv

        in this case the `partitions` will be returned as:

        .. code-block::

            {'as_of_date': ['2019-09-10'], 'location': ['eu', 'us]}
        """
        if hasattr(self, 'organisation_short_code'):
            # Code that goes via S3 proxy requires that the dataset attributes contain the
            # `organisation_short_code`. A bug in the Catalogue means that the
            # `organisation_short_code` is not yet returned if the user gets a dataset by this
            # code path:
            #
            #   client.packages.get(name='example')
            #   datasets = packageDetails.datasets()
            #
            #   for key, dataset in datasets.items():
            #       dataset.metadata()
            #
            # so we need this if-statement until the Catalogue team release a fix to production.

            partitions: Dict[str, Set] = defaultdict(set)
            all_splits = (
                _get_partitions_in_filepath(path) for path in self.__list()
            )

            # Accumulate the values for each partition in a set, which will only
            # keep unique values.
            for splits in all_splits:
                for k, v in splits:
                    partitions[k].add(v)

            # We want to return the values as a sorted list.
            return {k: sorted(v) for k, v in partitions.items()}
        else:
            # Old code. This can only be removed when the Catalogue fixes their dataset responses
            # in production to return the organisation short code for all dataset calls.
            response = self._client.session.get(
                urljoin(
                    self._client._environment.consumption,
                    consumption_urls.consumption_partitions.format(id=self.id)
                )
            )

            return response.json()["data"]["attributes"]["partitions"]

    def list(
        self,
        request_id=None,
        partitions: Optional[Union[str, List[str]]] = None,
        filter_path: Optional[str] = None,
        absolute_path: bool = True,
        skip_hidden_files: bool = True,
    ) -> List[str]:
        """
        List all the paths to files in the dataset. Calls go via the
        Datalake's S3 proxy, so the returned paths will be returned in the
        style of the S3 proxy, so in the pattern
        `s3://organisation_short_code/dataset_short_code/<path>`.
        S3 proxy does not return direct paths to the real location on S3 as
        they may be sensitive.

        The list will filter out paths where:
        * The size is zero bytes.
        * The name of any partition within the path is prefixed with a `.` or
        a `_` as that means it is intended as a hidden file.
        * The name of any partition within the path is exactly the word
        `metadata`.
        * It is a directory rather than a file.
        * Files with the wrong extension for the dataset's type i.e for
        a .parquet dataset we will return only .parquet files, zero .csv
        files.

        This list will filter based on `load_type` of the dataset. When the
        dataset's `load_type` is `Incremental Load` then files will be listed
        from all of the `as_of_date` partitions on S3. When the
        dataset's `load_type` is `Full Load` then files will be listed only
        the most recent `as_of_date` partition on S3.
        Please see the support library documentation for more information
        about how the `load type` affects the data you can access in a dataset:
        https://supportlibrary.ihsmarkit.com/display/DLSL/Exploring+Datasets#ExploringDatasets-Howthe%60loadtype%60affectsthedatayoucanaccessinadataset

        Parameters
        ----------
        :param str request_id: Optional. Automatically generated value
            used by our logs to correlate a user journey across several
            requests and across multiple services.

        :param List[str] partitions: Optional. A list of filters (partitions)
            to apply to the file paths on S3.

            This allows the user to filter the files they want to read from
            before they  have to read any data. This will reduce the amount
            of redundant data that is transferred over the network and
            reduce the amount of memory required on the user's machine.

            For any dataset, the paths on S3 may contain
            partitions that are key-value pairs in the format
            `<key>=<value>`, for example in the path:

                s3://example/as_of_date=2020-01-01/country=US/file.extension

            there are the partitions:

            * `country=US`
            * `as_of_date=2020-01-01`.

            You can find the available list of partitions for a dataset by
            calling the dataset's `.partitions()` endpoint.

            You can specify a list partitions you want to get data from by
            passing in a list in the format `<key><operator><value>`. Please
            note that you should not include any extra quotation marks around
            the value, so `as_of_date=2020-01-01` is valid but
            `as_of_date='2020-01-01'` is invalid.

            The supported operators are: `<=`, `<`, `=`, `>`, `>=` and
            `!=`.

            For a partition where the key is named `as_of_date`, the value
            is treated as a Python datetime object, so comparisons are done
            between dates. This means that the `>=` and `<=` operators can
            be used to get data in a date range where more than one
            `as_of_date` exists. For example, to get data from only dates in
            January 2020:

            .. code-block:: python

                dataset.list(partitions=[
                    'as_of_date>=2020-01-01',
                    'as_of_date<=2020-01-31',
                ])

            Other key names will have their values treated as string, so the
            operators will perform Python string comparisons.

            .. code-block:: python

                dataset.list(partitions=[
                    'country=US'
                ])

            A list of partitions is applied as a logical AND. So specifying
            `as_of_date` and `country` will match only those paths that
            match both `as_of_date` AND `country`.

            .. code-block:: python

                dataset.list(partitions=[
                    'as_of_date>=2020-01-01',
                    'country=US',
                ])

        :param str filter_path: Optional. If provided only a subpath matching
            the filter_path will be matched. This is less flexible than using
            the `partitions` parameter, so we recommend you pass in
            `partitions` instead. The `partitions` can deal with the
            partitions in the path being in any order, but filter_path relies
            on a fixed order of partitions that the user needs to know ahead
            of time.

            Example usage to get all paths that start with the `as_of_date`
            2020. Note this is a string comparison, not a datetime comparison.

            .. code-block:: python

                dataset.list(filter_path='as_of_date=2020')

        :param bool absolute_path: True (default) for returning an
            absolute path to the path on the S3 proxy. False to return a
            relative path (this is useful if using a TransferManager).

        :param bool skip_hidden_files: True (default) skips files that have
            been uploaded to S3 by Spark jobs. These usually start with a
            `.` or `_`.
        """

        return self.__list(
            request_id=request_id,
            partitions=partitions,
            filter_path=filter_path,
            absolute_path=absolute_path,
            skip_hidden_files=skip_hidden_files,
        )

    def __list(
            self,
            request_id=None,
            partitions: Optional[Union[str, List[str]]] = None,
            filter_path: Optional[str] = None,
            absolute_path: bool = True,
            skip_hidden_files: bool = True,
    ) -> List[str]:
        """
        List all the paths to files in the dataset.

        This private function can be called internally by other functions
        such as `partitions` and `dask_dataframe`. This is private so that
        when there is an exception and aspects lists the public methods that
        we do not spam the logs.

        :param request_id:
        :param partitions:
        :param filter_path:
        :param absolute_path:
        :param skip_hidden_files:
        :return: List all the paths to files in the dataset.
        """
        if filter_path and partitions:
            raise ValueError(
                'Both `partitions` and `filter_path` are set. Please only '
                'provide one of these parameters.'
            )

        if type(partitions) == str:
            partitions = [partitions]

        return self._DatasetModel__list(
            request_id=request_id,
            custom_filter_func=partial(self.__custom_filter, partitions=partitions),
            filter_path=filter_path,
            absolute_path=absolute_path,
            skip_hidden_files=skip_hidden_files
        )

    def download(
        self,
        destination_path: str,
        flatten: Optional[bool] = False,
        filter_path: Optional[str] = None,
        partitions: Optional[Union[str, List[str]]] = None,
    ) -> List[str]:
        """
        Downloads the original dataset files from the data lake to local copy
        destination of your choice.

        The flatten parameter retains only the file name, and places the
        files in the directory specified, else the files will be downloaded
        matching the directory structure as housed on the data lake.

        The filter path and partitions parameters specify that only a subset
        of the S3 path should be downloaded.

        Parameters
        ----------
        :param destination_path: required. The path on the system, where the
            files should be saved. Must be a directory, if doesn't exist, will
            be created.

        :param bool flatten: The flatten parameter default behaviour
            (flatten=False) will allow the user to specify that they would
            like to keep the underlying folder structure when writing the
            downloaded files to disk.
            When “flatten” is set to True the
            folder structure and download all data to the specified path in
            and retains only the original file name. This is not useful if the
            file names are identical for multiple files.

        :param str filter_path: Optional. If provided only a subpath matching
            the filter_path will be matched. This is less flexible than using
            the `partitions` parameter, so we recommend you pass in
            `partitions` instead. The `partitions` can deal with the
            partitions in the path being in any order, but filter_path relies
            on a fixed order of partitions that the user needs to know ahead
            of time.

            Example usage to get all paths that start with the `as_of_date`
            2020. Note this is a string comparison, not a datetime comparison.

            .. code-block:: python

                dataset.download(filter_path='as_of_date=2020')

        :param List[str] partitions: Optional. A list of filters (partitions)
            to apply to the file paths on S3.

            This allows the user to filter the files they want to read from
            before they  have to read any data. This will reduce the amount
            of redundant data that is transferred over the network and
            reduce the amount of memory required on the user's machine.

            For any dataset, the paths on S3 may contain
            partitions that are key-value pairs in the format
            `<key>=<value>`, for example in the path:

                s3://example/as_of_date=2020-01-01/country=US/file.extension

            there are the partitions:

            * `country=US`
            * `as_of_date=2020-01-01`.

            You can find the available list of partitions for a dataset by
            calling the dataset's `.partitions()` endpoint.

            You can specify a list of partitions you want to get data from by
            passing in a list in the format `<key><operator><value>`. Please
            note that you should not include any extra quotation marks around
            the value, so `as_of_date=2020-01-01` is valid but
            `as_of_date='2020-01-01'` is invalid.

            The supported operators are: `<=`, `<`, `=`, `>`, `>=` and
            `!=`.

            For a partition where the key is named `as_of_date`, the value
            is treated as a Python datetime object, so comparisons are done
            between dates. This means that the `>=` and `<=` operators can
            be used to get data in a date range where more than one
            `as_of_date` exists. For example, to get data from only dates in
            January 2020:

            .. code-block:: python

                dataset.download(partitions=[
                    'as_of_date>=2020-01-01',
                    'as_of_date<=2020-01-31',
                ])

            Other key names will have their values treated as string, so the
            operators will perform Python string comparisons.

            .. code-block:: python

                dataset.download(partitions=[
                    'country=US'
                ])

            A list of partitions is applied as a logical AND. So specifying
            `as_of_date` and `country` will match only those paths that
            match both `as_of_date` AND `country`.

            .. code-block:: python

                dataset.download(partitions=[
                    'as_of_date>=2020-01-01',
                    'country=US',
                ])

        :return: the list of the files that were downloaded successfully. Any
            failures will be printed.


        :example:

            Downloading without flatten:

            .. code-block:: python

                dataset.download('./local/path/')

                [
                  'as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
                  'as_of_date=2019-09-11/type=full/StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
                ]

        :example:

            Downloading with ``filter_path``

            .. code-block:: python

                dataset.download(
                    './local/path/', filter_path='as_of_date=2019-09-10/'
                )
                [
                  'as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
                ]


        :example:

            When flatten = True, we remove the s3 structure. Example:

            Example output for new behaviour:

            .. code-block:: python

                dataset.download('./local/path/', flatten=True)
                [
                  'StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
                  'StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
                ]


        """

        if filter_path and partitions:
            raise ValueError(
                'Both `partitions` and `filter_path` are set. Please only '
                'provide one of these parameters.'
            )

        if type(partitions) == str:
            partitions = [partitions]

        return self._DatasetModel__download(
            destination_path=destination_path,
            flatten=flatten,
            filter_path=filter_path,
            custom_filter_func=partial(self.__custom_filter, partitions=partitions),
            skip_hidden_files=True
        )

    def dataframe(
        self,
        nrows: Optional[int] = None,
        partitions: Optional[Union[str, List[str]]] = None,
        raise_: bool = True,
        use_compression: bool = False
    ) -> pandas.DataFrame:
        """
        Return the data from the files in the latest instance of the dataset
        as a pandas DataFrame.

        WARNING: The dependency Pandas 1.0.4 has an issue reading parquet due
        to an pandas library mismatch. This was fixed in Pandas 1.0.5.

        We currently support .csv and .parquet as our data file formats. The
        data files in the latest instance could all be .csv format or all be
        .parquet format. If there is a mix of .csv and .parquet or some other
        file format then we will not be able to parse the data and will
        return an error message.

        :param int nrows: Optional. The max number of rows to return.
            We use the nrows parameter to limit the amount of rows that are
            returned, otherwise for very large dataset it will take a long time
            or you could run out of RAM on your machine!
            If you want all of the rows, then leave this parameter set to the
            default None.

        :param List[str] partitions: Optional. A list of filters (partitions)
            to apply to the file paths on S3.

            This allows the user to filter the files they want to read from
            before they  have to read any data. This will reduce the amount
            of redundant data that is transferred over the network and
            reduce the amount of memory required on the user's machine.

            For any dataset, the paths on S3 may contain
            partitions that are key-value pairs in the format
            `<key>=<value>`, for example in the path:

                s3://example/as_of_date=2020-01-01/country=US/file.extension

            there are the partitions:

            * `country=US`
            * `as_of_date=2020-01-01`.

            You can find the available list of partitions for a dataset by
            calling the dataset's `.partitions()` endpoint.

            You can specify a list of partitions you want to get data from by
            passing in a list in the format `<key><operator><value>`. Please
            note that you should not include any extra quotation marks around
            the value, so `as_of_date=2020-01-01` is valid but
            `as_of_date='2020-01-01'` is invalid.

            The supported operators are: `<=`, `<`, `=`, `>`, `>=`.

            For a partition where the key is named `as_of_date`, the value
            is treated as a Python datetime object, so comparisons are done
            between dates. This means that the `>=` and `<=` operators can
            be used to get data in a date range where more than one
            `as_of_date` exists. For example, to get data from only dates in
            January 2020:

            .. code-block:: python

                dataset.dataframe(partitions=[
                    'as_of_date>=2020-01-01',
                    'as_of_date<=2020-01-31',
                ])

            Other key names will have their values treated as string, so the
            operators will perform Python string comparisons.

            .. code-block:: python

                dataset.dataframe(partitions=[
                    'country=US'
                ])

            A list of partitions is applied as a logical AND. So specifying
            `as_of_date` and `country` will match only those paths that
            match both `as_of_date` AND `country`.

            .. code-block:: python

                dataset.dataframe(partitions=[
                    'as_of_date>=2020-01-01',
                    'country=US',
                ])

        :param bool raise_: Optional. Raise exception if the dataframe stream
            stopped prematurely

        :param use_compression: Optional. Whether the response from the
        backend should use compression. Setting to false should result
        in a faster initial response before the streaming begins.

        :example:

            Basic usage:

            .. code-block:: python

                    dataframe = dataset.dataframe()

        :example:

            Dataframe filtered by partition with nrows (partitions can be
            fetched using the `dataset.partitions()` endpoint:

            .. code-block:: python

                    dataframe = dataset.dataframe(
                        nrows=1000,
                        partitions=["as_of_date=2017-03-07"]
                    )
        """

        self._check_access()

        params = {}

        if nrows is not None:
            params['filter[nrows]'] = nrows

        if partitions is not None:
            if type(partitions) == str:
                partitions = [partitions]

            params['filter[partitions]'] = partitions

        dataframe_url = urljoin(
            self._client._environment.consumption,
            consumption_urls.consumption_dataframe.format(id=self.id)
        )

        headers = {}
        if not use_compression:
            headers['Accept-Encoding'] = 'identity;q=0'

        response = self._client.session.get(
            dataframe_url, stream=True,
            params=params,
            headers=headers,
        )

        try:
            # native_file only reads until end of dataframe
            # the rest of the stream has to be read from response raw.
            native_file = pyarrow.PythonFile(response.raw, mode='rb')

            # Now native_file "contains the complete stream as an in-memory
            # byte buffer. An important point is that if the input source
            # supports zero-copy reads (e.g. like a memory map,
            # or pyarrow.BufferReader), then the returned batches are also
            # zero-copy and do not allocate any new memory on read."
            reader = pyarrow.ipc.open_stream(native_file)

            dataframe = reader.read_pandas()
        except (
                ArrowInvalid,
                http.client.IncompleteRead,
                ProtocolError
        ) as ex:

            message = 'Sorry, the dataframe you are trying to read is too ' \
                      'large to fit into memory. Please consider one of ' \
                      'these alternatives:' \
                      '\n\n1. Run the same code on a machine that has more ' \
                      'RAM, for example a Jupyter Notebook hosted on an ' \
                      'AWS environment.' \
                      '\n2. Use a big data tool such as Spark, Dask or ' \
                      'similar to read the data via our S3 proxy. Unlike ' \
                      'Pandas, these big data tools are able to apply some ' \
                      'operations such as filtering without having to hold ' \
                      'all of the data into memory. See ' \
                      'https://supportlibrary.ihsmarkit.com/display/DLSL/' \
                      'Using+the+S3+Proxy+with+Big+Data+tooling' \
                      f'\n\nOriginal exception: {ex}'

            raise ArrowInvalid(message) from ex

        # The pyarrow buffered stream reader stops once it
        # reaches the end of the IPC message. Afterwards we
        # get the rest of the data which contains the summary
        # of what we've downloaded including an error message.
        last_packet = response.raw.read()
        summary = json.loads(last_packet)

        if summary['status'] >= 400:
            exception = DataframeStreamingException(
                summary, dataframe_url, response=response,
            )

            # Optionally ignore bad data
            if raise_:
                raise exception
            else:
                warnings.warn(
                    str(exception),
                    UserWarning
                )

        return dataframe

    def contents(self):
        """
        Print IDs for all the instances in this dataset.

        Example output:

            INSTANCE 1111aaa-11aa-11aa-11aa-111111aaaaaa
        """
        for p in self.instances.all():
            print(str(p))

    def dictionary(self) -> List[dict]:
        """
        List of dictionaries for each field in the dataset's dictionary as
        returned by the Catalogue.

        If there are multiple historical values for the dataset's `dictionary`
        in the Catalogue, then you will see more than one item in the list,
        otherwise there will only be a single item in the list containing a
        dictionary of the current fields.

        Field names may include:

        * name
        * type
        * nullable
        * description
        * comment
        * validation
        * short_name
        * sample_value
        * is_derived
        * metadata

        Example usage:

        .. code:: python

            dataset = client.datasets.get('example-dataset-short-code')
            print(dataset.dictionary())

        Example output:

        .. code:: python

            [
              {
                'name': 'some-name',
                'type': 'String',
                'nullable': True,
                'description': 'some-description'
              }
            ]

        :return: List of each field in the dataset.
        """

        if self.fields_metadata is None:
            # we have to do two calls
            # to get the latest dictionary id for the dataset
            try:
                response = self._client.session.get(
                    dataset_urls.dictionary_by_dataset_lastest.format(id=self.id)
                ).json()
            except:
                print("There is no current dictionary available.")
                return []

            # followed by the fields for the dictionary...

            self.fields_metadata = DictionaryModel(
                {'id': response["data"]["id"], 'attributes': {}},
                client=self._client
            ).fields

            def subdict(field_dict):
                return {k: v for k, v in field_dict.items()}

            self.fields_metadata = list(
                map(lambda field: subdict(field),
                    self.fields_metadata)
            )

        return self.fields_metadata

    def info(self):
        fields = self.dictionary()

        df = pandas.DataFrame(fields)
        if df.shape[1] > 0:
            df["type"] = df.apply(
                lambda row: row["type"] + (" (Nullable)" if row["nullable"] else ""),
                axis=1)
            df = df[["name", "type"]]

            print(tabulate(df, showindex=False, headers=df.columns))
        else:
            print("No columns/info available.")


log_public_functions_calls_using(
    [analytics_decorator, logging_decorator],
    class_fields_to_log=['dataset_id']
)(StructuredDatasetModel)



#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import contextlib
import os
import shutil
import ntpath
from typing import Generator
from io import BufferedReader
from urllib.parse import urljoin, urlparse

from dli.client.aspects import analytics_decorator, logging_decorator
from dli.client.components.urls import consumption_urls
from dli.models import log_public_functions_calls_using, AttributesDict


def flatten_s3_file_path(s3_path):
    path = urlparse(s3_path)[2]  # [scheme, netloc, path, ...]
    head, tail = ntpath.split(path)
    file = tail or ntpath.basename(head)
    return file


def get_or_create_os_path(s3_path: str, to: str, flatten: bool) -> str:
    if flatten:
        destination_key = flatten_s3_file_path(s3_path)
    else:
        destination_key = urlparse(s3_path).path.lstrip('/')

    to_path = os.path.join(
        to, os.path.normpath(destination_key)
    )

    to_path = os.path.abspath(to_path)

    if len(to_path) > 259 and os.name == 'nt':
        raise Exception(f"Apologies {s3_path} can't be downloaded "
                        f"as the file name would be too long. You "
                        f"may want to try calling again with "
                        f"Instance.download(flatten=True), which "
                        f"will put the file in a directory of your choice")

    else:
        directory, _ = os.path.split(to_path)
        os.makedirs(directory, exist_ok=True)

    return to_path


class FileModel(AttributesDict):

    @contextlib.contextmanager
    def open(self) -> Generator[BufferedReader, None, None]:
        """
        Context manager that yields a file like object.

        :example:
            
            .. code-block::

                with file.open() as f:
                    print(f.read())
                    
        
        .. note::
            
            The return type is a BufferedReader. This is the
            raw HTTP stream. This means that unlike usual file
            objects you can no seek it, it can only be read
            in order. So save to a file or buffer if you wish
            to manipulate the file out of order.

        """
        response = self._client.session.get(
            urljoin(
                self._client._environment.consumption,
                consumption_urls.consumption_download.format(
                    id=self.datafile_id,
                    path=self.path
                )
            ),
            stream=True
        )
        # otherwise you get raw secure
        response.raw.decode_content = True
        yield response.raw
        response.close()

    def download(self, to='./', flatten=False) -> str:
        """
        Download the files for the instance, then return a list of the file
        paths that were written.

        :param str to: The path on the system, where the files
            should be saved. must be a directory, if doesn't exist, will be
            created.

        :param bool flatten: The default behaviour (=False) is to use the s3
            file structure when writing the downloaded files to disk.
            When flatten = True, we remove the s3 structure.

        :return: Directory to path of the downloaded files.

        :example:

            Downloading without flatten:

            .. code-block:: python

                >>> dataset.instances.lastest().files()[0].download('./local/path/')
                './local/path/as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',


        :example:

            When flatten = True, we remove the s3 structure. Example:

            Example output for new behaviour:

            .. code-block:: python

                >>> dataset.instances.lastest().files()[0].download('./local/path/', flatten=True)
                './local/path/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',

        """
        to_path = get_or_create_os_path(
            s3_path=self.path, to=to, flatten=flatten
        )

        print(f'Downloading {self.path} to: {to_path}...')

        with self.open() as download_stream:
            with open(to_path, 'wb') as target_download:
                # copyfileobj is just a simple buffered
                # file copy function with some sane
                # defaults and optimisations.
                shutil.copyfileobj(
                    download_stream, target_download
                )
                print(f'Completed download to: {to_path}.')

        return to_path


log_public_functions_calls_using(
    [analytics_decorator, logging_decorator],
    class_fields_to_log=['datafile_id', 'path']
)(FileModel)



from datetime import datetime
from typing import List
from unittest.mock import Mock, PropertyMock

import pytest
from freezegun import freeze_time
from moto import mock_s3

from dli.aws import create_refreshing_session, _meets_partition_params, \
    match_partitions, get_most_recent_common_prefix

path_with_as_of_date = 'prefix/a=2/b=4/c=6/as_of_date=2020-01-23/file.ext'
path_with_as_of_year_month = 'prefix/a=2/b=4/c=6/' \
                             'as_of_year_month=2020-01/file.ext'
paths_and_sizes = [
    'prefix/as_of_date=2020-01-01/file0.ext',
    'prefix/as_of_date=2020-01-02/file1.ext',
    'prefix/as_of_date=2020-01-03/file2.ext',
    'prefix/as_of_date=2020-01-03/file3.ext',
]
paths_without_as_of = [
    'prefix/file0.ext',
    'prefix/file1.ext',
]


@mock_s3
def test_session_refreshes():
    dli_client = Mock()
    # This is how you give an attribute a side effect
    auth_key = PropertyMock(return_value='abc')
    type(dli_client.session).auth_key = auth_key 

    dli_client.session.token_expires_on = datetime(2000, 1, 2, 0, 0, 0)
    session = create_refreshing_session(dli_client.session).resource('s3')

    with freeze_time(datetime(2000, 1, 1, 0, 0, 0)):
        assert auth_key.call_count == 1
        session.meta.client.list_buckets()
        assert auth_key.call_count == 1
        session.meta.client.list_buckets()
        assert auth_key.call_count == 1

    with freeze_time(datetime(2000, 1, 3, 0, 0, 0)):
        dli_client.session.token_expires_on = datetime(2000, 1, 4, 0, 0, 0)
        assert auth_key.call_count == 1
        session.meta.client.list_buckets()
        assert auth_key.call_count == 2
        session.meta.client.list_buckets()
        assert auth_key.call_count == 2


@pytest.mark.parametrize(
    'path,partition_query',
    [
        [
            path_with_as_of_date,
            [{'partition': 'a', 'operator': '>', 'value': '3'}],
        ],

        [
            path_with_as_of_date,
            [{'partition': 'c', 'operator': "<", 'value': '5'}],
        ],

        [
            path_with_as_of_date,
            [{'partition': 'as_of_date', 'operator': '<',
              'value': '2020-01-23'}],
        ],
        [
            path_with_as_of_year_month,
            [{'partition': 'as_of_year_month', 'operator': '<',
              'value': '2020-01'}],
        ],

        [
            path_with_as_of_date,
            [{'partition': 'as_of_date', 'operator': '>',
              'value': '2020-01-23'}],
        ],
        [
            path_with_as_of_year_month,
            [{'partition': 'as_of_year_month', 'operator': '>',
              'value': '2020-01'}],
        ],
    ],
)
def test_fails_partition_params(path: str, partition_query: List):
    assert not _meets_partition_params(path, partition_query)


@pytest.mark.parametrize(
    'path,partition_query',
    [
        [
            path_with_as_of_date,
            [
                {'partition': 'as_of_date', 'operator': '=',
                 'value': '2020-01-23'}
            ]
        ],
        # Left starts with space
        [
            path_with_as_of_date,
            [{'partition': ' as_of_date', 'operator': '=',
              'value': '2020-01-23'}
             ]
        ],
        # Left ends with space
        [
            path_with_as_of_date,
            [{'partition': 'as_of_date ', 'operator': '=',
              'value': '2020-01-23'}
             ]
        ],
        # Left starts and ends with space
        [
            path_with_as_of_date,
            [{'partition': ' as_of_date ', 'operator': '=',
              'value': '2020-01-23'}
             ]
        ],
        # Right starts with space
        [
            path_with_as_of_date,
            [{'partition': 'as_of_date', 'operator': '=',
              'value': ' 2020-01-23'}
             ]
        ],
        # Right ends with space
        [
            path_with_as_of_date,
            [{'partition': 'as_of_date', 'operator': '=',
              'value': '2020-01-23 '}
             ]
        ],
        # Right starts and ends with space
        [
            path_with_as_of_date,
            [{'partition': 'as_of_date', 'operator': '=',
              'value': ' 2020-01-23 '}
             ]
        ],

        # Left starts with space
        [
            path_with_as_of_year_month,
            [{'partition': ' as_of_year_month', 'operator': '=',
              'value': '2020-01'}
             ]
        ],
        # Left ends with space
        [
            path_with_as_of_year_month,
            [{'partition': 'as_of_year_month ', 'operator': '=',
              'value': '2020-01'}
             ]
        ],
        # Left starts and ends with space
        [
            path_with_as_of_year_month,
            [{'partition': ' as_of_year_month ', 'operator': '=',
              'value': '2020-01'}
             ]
        ],
        # Right starts with space
        [
            path_with_as_of_year_month,
            [{'partition': 'as_of_year_month', 'operator': '=',
              'value': ' 2020-01'}
             ]
        ],
        # Right ends with space
        [
            path_with_as_of_year_month,
            [{'partition': 'as_of_year_month', 'operator': '=',
              'value': '2020-01 '}
             ]
        ],
        # Right starts and ends with space
        [
            path_with_as_of_year_month,
            [{'partition': 'as_of_year_month', 'operator': '=',
              'value': ' 2020-01 '}
             ]
        ],
    ]
)
def test_meets_date_supported_partition_params(
        path: str,
        partition_query: List
):
    assert _meets_partition_params(path, partition_query)


@pytest.mark.parametrize(
    'path,partition_query', [
        [
            path_with_as_of_date,
            [{'partition': 'a', 'operator': '<', 'value': '3'}],
        ],

        [
            path_with_as_of_date,
            [{'partition': 'b', 'operator': '=', 'value': '4'}],
        ],

        [
            path_with_as_of_date,
            [{'partition': 'c', 'operator': ">", 'value': '5'}],
        ],

        [
            path_with_as_of_date,
            [{'partition': 'as_of_date', 'operator': '=',
              'value': '2020-01-23'}],
        ],
        [
            path_with_as_of_year_month,
            [{'partition': 'as_of_year_month', 'operator': '=',
              'value': '2020-01'}],
        ],

        # Combination of all filters.
        [
            path_with_as_of_date,
            [{'partition': 'a', 'operator': '<', 'value': '3'},
             {'partition': 'b', 'operator': '=', 'value': '4'},
             {'partition': 'c', 'operator': ">", 'value': '5'},
             {'partition': 'as_of_date', 'operator': '=',
              'value': '2020-01-23'}],
        ],
        [
            path_with_as_of_year_month,
            [{'partition': 'a', 'operator': '<', 'value': '3'},
             {'partition': 'b', 'operator': '=', 'value': '4'},
             {'partition': 'c', 'operator': ">", 'value': '5'},
             {'partition': 'as_of_year_month', 'operator': '=',
              'value': '2020-01'}],
        ],
    ],
)
def test_meets_partition_params(path: str, partition_query: List):
    assert _meets_partition_params(path, partition_query)

@pytest.mark.parametrize(
    'partition_query', [
        [
            {'partition': 'd', 'operator': '<', 'value': '5'}
        ]
    ],
)
def test_fails_unknown_partition_params(
    partition_query: List
):
    assert not _meets_partition_params(path_with_as_of_date, partition_query)



@pytest.mark.parametrize(
    'path,partitions', [
        [
            path_with_as_of_date, ['a<3']
        ],
        [
            path_with_as_of_date, ['b=4']
        ],
        [
            path_with_as_of_date, ['c>5']
        ],
        [
            path_with_as_of_date, ['as_of_date=2020-01-23']
        ],
        [
            path_with_as_of_year_month, ['as_of_year_month=2020-01']
        ],
        # Combination of all filters.
        [
            path_with_as_of_date,
            ['a<3', 'b=4', 'c>5', 'as_of_date=2020-01-23']
        ],
        [
            path_with_as_of_year_month,
            ['a<3', 'b=4', 'c>5', 'as_of_year_month=2020-01']
        ],
        [
            'prefix/export_region=NorthAmerica/file.ext',
            ['export_region=NorthAmerica']
        ],
        [
            'prefix/export_region=North_America/file.ext',
            ['export_region=North_America']
        ],
        [
            'prefix/export_region=North America/file.ext',
            ['export_region=North America']
        ],
        [
            'prefix/export region=North America/file.ext',
            ['export region=North America']
        ],
    ],
)
def test_match_partitions(path: str, partitions: List[str]):
    assert match_partitions(path, partitions)


@pytest.mark.parametrize('path,partition_query', [
    [path_with_as_of_date, [{
        'partition': 'as_of_date',
        'operator': '=',
        'value': '202-01-23'
    }]],
    [path_with_as_of_year_month, [{
        'partition': 'as_of_year_month',
        'operator': '=',
        'value': '202-01'
    }]],
])
def test_raises_error_with_malformed_date(path: str, partition_query: List):
    with pytest.raises(ValueError):
        _meets_partition_params(path, partition_query)


def test_false_when_as_of_date_is_invalid():
    partition_query = [{
        'partition': 'as_of_date',
        'operator': '=',
        'value': '2020-01-23'
    }]

    invalid_path = 'b=4/c=6/as_of_date=ABC/file.ext'

    assert not _meets_partition_params(invalid_path, partition_query)


def test_load_type_full_returns_data_from_most_recent_as_of_date():
    assert get_most_recent_common_prefix(
        common_prefixes=paths_and_sizes,
    ) == paths_and_sizes[-2:]


def test_load_type_full_but_zero_paths_returns_empty():
    assert get_most_recent_common_prefix(
        common_prefixes=[],
    ) == []


def test_load_type_full_all_paths_without_as_of_date_return_paths():
    assert get_most_recent_common_prefix(
        common_prefixes=paths_without_as_of,
    ) == paths_without_as_of


def test_load_type_full_some_paths_without_as_of_date_return_paths():
    assert get_most_recent_common_prefix(
        common_prefixes=paths_and_sizes + paths_without_as_of,
    ) == paths_and_sizes[-2:]



#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import logging
import warnings

import requests

from dli.client.components.urls import package_urls
from dli.client.components import SirenComponent, SirenAdapterResponse
from dli.client.exceptions import CatalogueEntityNotFoundException
from dli.client.utils import filter_out_unknown_keys, to_camel_cased_dict

logger = logging.getLogger(__name__)


class Package(SirenComponent):

    _KNOWN_FIELDS = {"name",
                     "description",
                     "keywords",
                     "topic",
                     "access",
                     "internalData",
                     "contractIds",
                     "termsAndConditions",
                     "derivedDataNotes",
                     "derivedDataRights",
                     "distributionNotes",
                     "distributionRights",
                     "internalUsageNotes",
                     "internalUsageRights",
                     "documentation",
                     "publisher",
                     "techDataOpsId",
                     "accessManagerId",
                     "managerId",
                     "intendedPurpose",
                     "isInternalWithinOrganisation"}
    """
    A mixin providing common package operations
    """

    def get_package(self, id=None, name=None):
        """
        Fetches package metadata for an existing package.

        :param str id: The id of the package.
        :param str name: The name of the package. We will perform a case
        sensitive exact match on this name in the database. If you want
        a non-exact match or a case insensitive match, then please consider
        using the `packages` call instead with the `search_term` parameter
        and an `ilike` matcher, for example:

        .. code:: python

            client.packages(search_term='name ilike example')

        :returns: A package instance
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                # Look up by package id
                package = client.get_package('my_package_id')
                # or
                package = client.get_package(id='my_package_id')

                # Alternatively look up by package name
                package = client.get_package(name='my_package')

        """

        warnings.warn(
            'This method is deprecated and is scheduled for removal in '
            'dli==1.14.0. '
            '\nTo search for a package, please use this call instead:'
            '\n'
            '\n    import dli'
            '\n    client = dli.connect()'
            '\n    # Return type: OrderedDict[str, PackageModel]'
            '\n    packages = client.packages(search_term=["name=<package_name>"])',
            PendingDeprecationWarning
        )

        if id is None and name is None:
            raise ValueError("Either package id or name must be specified to look up package")

        return self._get_package(package_id=id, name=name)

    def get_package_datasets(self, package_id, count=100):
        """
        Returns a list of all datasets registered under a package.

        :param str package_id: The id of the package.
        :param int count: Optional. Count of datasets to be returned. Defaults to 100.

        :returns: List of all datasets registered under the package.
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                datasets = client.get_package_datasets(
                    package_id,
                    count=100
                )
        """
        response = self.session.get(
            package_urls.v2_package_datasets.format(id=package_id),
            params={'page_size': count}
        )

        return self._DatasetFactory._from_v2_list_response(response.json())

    @staticmethod
    def get_default_package_terms_and_conditions(organisation_name: str):
        """
        Returns a string representing the default Terms And Conditions for packages created in DataLake.

        :returns: The default DataLake Terms And Conditions
        :rtype: str
        """
        if organisation_name == 'IHS Markit':
            return ('By submitting this Data request and checking the "Accept Terms and Conditions" '
                'box, you acknowledge and agree to the following:\n'
                '\n'
                '* To promptly notify the relevant Access Manager/Producer of your intended use '
                'of the Data;\n'
                '* To obtain the terms and conditions relevant to such use for such Data from '
                'the Producer;\n'
                '* To distribute such terms and conditions to each member of your '
                'Consumer Group who may use the Data;\n'
                '* To use the Data solely for such intended use, subject to such terms and '
                'conditions;\n'
                '* To ensure that the Data is only accessed by members of your Consumer Group, '
                'and only used by such members for such intended use, subject to such terms and '
                'conditions;\n'
                '* To adhere to any additional requests of Producer with respect to the Data '
                '(including but not limited to ceasing use of the Data and deleting the Data, '
                'and ensuring other members of the Consumer Group do so, upon revocation of your '
                'license by Producer).\n'
                '\n'
                'Please refer to the <a href="/terms-of-use" target="_blank">EULA</a> for any '
                'defined terms used above. '
                'The <a href="/terms-of-use" target="_blank">EULA</a> '
                'is the document you agreed to adhere to by accessing the Lake.')
        else:
            return ''

    #
    # Private functions
    #
    def _get_package(self, **kwargs):
        vals = []

        if 'package_id' in kwargs and kwargs['package_id'] is not None:
            vals = self.packages(
                search_term=[f"id={kwargs['package_id']}"]).values()

        elif 'name' in kwargs and kwargs['name'] is not None:
            vals = self.packages(
                search_term=[f"name={kwargs['name']}"]).values()

        else:
            raise ValueError("Either package id or name must be specified to look up package")

        vals = list(vals)

        if len(vals) == 1:
            if hasattr(vals[0], "id"):
                # Add an alias of the `id` attribute named `package_id`.
                vals[0].package_id = vals[0].id

            return vals[0]
        elif len(vals) < 1:
            r = requests.Response()
            r.status_code = 404
            r.request = (lambda: True)
            r.request.method = f"Package {kwargs['package_id'] if kwargs.get('package_id') else kwargs['name']}"
            r.request.url = ""
            raise CatalogueEntityNotFoundException(
                message=f"Package "
                f"{kwargs['package_id'] if kwargs.get('package_id') else kwargs['name']}",
                response=r
            )
        else:
            return vals

    @staticmethod
    def _validate_fields(fields):
        if 'termsAndConditions' in fields:
            Package._validate_terms_and_conditions(fields.get("termsAndConditions"))

    @staticmethod
    def _validate_terms_and_conditions(field):
        if not field or not field.strip():
            raise ValueError("Terms and conditions must be defined "
                             "and be non empty, non blank string")



#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import os
from collections import OrderedDict
from typing import Dict
from functools import partial
import requests

from dli.client.exceptions import CatalogueEntityNotFoundException
from dli.models.paginator import Paginator
from dli.client.components.urls import dataset_urls, identity_urls
from dli.models.dataset_model import DatasetModel
from dli.modules.package_module import PackageModule


class DatasetModule:

    def __call__(self, search_term=None, only_mine=False, warn=False, page_size=5000) \
            -> Dict[str, DatasetModel]:
        """
        See datasets.

        :Example:

            Get all datasets (including those you do not yet have access to):

            .. code:: python

                client.datasets()

        :Example:

            Get **only** datasets you have access to and
            can read:

            .. code:: python

                client.datasets(only_mine=True)

        :Example:

            Search for datasets using `ilike`. This will find fields that are
            similar to value. This is recommended over using `like` as it
            performs case insensitive matching. Example usages `ilike example`.

            Some examples of how the logic matches in the database:

            'example' ILIKE 'example'    true

            'example' ILIKE 'ex%'     true

            'example' ILIKE 'c'      false

            .. code:: python

                # Exact match `example` (case sensitive).
                client.datasets(search_term='name=example')

                # Match `example` (case insensitive).
                client.datasets(search_term='name ilike example')

                # Match `example` (case insensitive) followed by any characters.
                client.datasets(search_term='name ilike example%')

                # Match `example` (case insensitive) in the middle of any characters.
                client.datasets(search_term='name ilike %example%')

        :Example:

            Search for datasets with multiple search terms:

            .. code:: python

                client.datasets(
                    search_term=[
                        'name=example',
                        'manager_id=example'
                    ]
                )

        :Example:

            Search for datasets with name and only_mine:

            .. code:: python

                client.datasets(
                    search_term='name=example',
                    only_mine=True,
                )


        :param search_term: search term for dataset. Can filter using a simple
            query language. E.G "field=value", "field>value", "field ilike value".
            Can pass in multiple values using a list of strings.

        :param bool only_mine: Specify whether to collect datasets only
        accessible to you (True) or to discover packages that you may
        want to discover but may not have access to (False).

        :param int page_size: Optional. Default 5000. Used to control the size
        of internal calls from the SDK to the Catalogue. This parameter is
        only exposed to allow performance testing and changes for batch jobs.
        Customers should use the default setting.

        :returns: Ordered dictionary of ids to DatasetModel.
        :rtype: OrderedDict[id: str, DatasetModel]
        """

        filters = {}
        for x in PackageModule._filter_creation(search_term, only_mine):
            filters.update(x)

        p = Paginator(
            dataset_urls.v2_index,
            self._client._DatasetFactory,
            partial(self._client._DatasetFactory._from_v2_response_unsheathed, warn=warn),
            page_size=page_size,
            max_workers=5,
            filters=filters
        )

        results = [(v.short_code, v) for v in p]
        o = OrderedDict()
        for code, val in results:
            if code in o:
                o[code] = [o[code]]
                o[code].append(val)
            else:
                o[code] = val

        return o

    def get(self, short_code, organisation_short_code=None, page_size=5000) -> DatasetModel:
        """
        Returns a DatasetModel if it exists, and a user has access else None

        :param str short_code: The short code of the dataset to collect

        :param organisation_short_code:  The short code of the organisation
        for this dataset

        :param int page_size: Optional. Default 5000. Used to control the size
        of internal calls from the SDK to the Catalogue. This parameter is
        only exposed to allow performance testing and changes for batch jobs.
        Customers should use the default setting.

        :returns: Dataset model with matching short code.
        :rtype: DatasetModel
        """

        search_terms = [
            f"short_code={short_code}"
        ]

        if organisation_short_code is not None:
            search_terms.append(
                f"organisation_short_code={organisation_short_code}"
            )

        visible_orgs = self._client.session.get(
            identity_urls.orgs_visible_to_user
        ).json()

        mapping_to_sc = dict(map(
            lambda x: (x["id"],
                       x["attributes"]["short_code"]
                       ),
             visible_orgs["data"])
        )

        # todo - not going to work as expected since the orderdict wont have dups
        res = self._client.datasets(search_term=search_terms, only_mine=False, warn=True, page_size=page_size)
        res = res.get(short_code)

        if type(res) is list:
            if os.name != 'nt':
                org_codes=[f"\033[4m\033[92m\033[1morganisation_short_code=" \
                           f"{mapping_to_sc.get(x.organisation_id, x.organisation_id)}" \
                           f"\033[0;0m -> {x.short_code}"
                           for x in res]

                msg = f"\033[96mMultiple datasets have shortcode: \033[1m`{short_code}`\033[0;0m\n\nShortcodes are"\
                      f" only guaranteed to be unique per organisation, and you seem to "\
                      f" have access to multiple organisation's datasets, where there are "\
                      f"datasets bearing the same shortcode."\
                      f"\n"\
                      f"To resolve this, please specify the organisation_short_code in "\
                      f"the call to datasets.get(shortcode, organisation_short_code=ORG CODE)"\
                      f"\n\nThe available (accessible to you) organisations shortcodes "\
                      f"for this dataset short code are:\n" + ("\n".join(org_codes))
            else:
                org_codes = [f"organisation_short_code=" \
                             f"{mapping_to_sc.get(x.organisation_id, x.organisation_id)}" \
                             f" -> {x.short_code}"
                             for x in res]

                msg = f"Multiple datasets have shortcode: `{short_code}`\n\nShortcodes are" \
                      f" only guaranteed to be unique per organisation, and you seem to " \
                      f" have access to multiple organisation's datasets, where there are " \
                      f"datasets bearing the same shortcode." \
                      f"\n" \
                      f"To resolve this, please specify the organisation_short_code in " \
                      f"the call to datasets.get(shortcode, organisation_short_code=ORG CODE)" \
                      f"\n\nThe available (accessible to you) organisations shortcodes " \
                      f"for this dataset short code are:\n" + (
                          "\n".join(org_codes))


            raise Exception(
                msg
            )

            #todo - difficult here, really need to convert id to shortcode for org
            #todo - else isnt that useful a message as user now needs to discover these
        elif res:
            return res
        else:
            if self._client.strict:
                r = requests.Response()
                r.status_code = 404
                r.request = (lambda: True)
                r.request.method = f"No such dataset {short_code}"
                r.request.url = ""
                raise CatalogueEntityNotFoundException(
                    message=f"No such dataset {short_code}",
                    response=r
                )
            else:
                return None



import io
import json
import os
import random
import re
import tempfile
from contextlib import redirect_stdout
from functools import partial
from unittest.mock import MagicMock, call

import jwt
import numpy
import pandas
import pyarrow
import pytest

from dli.client.exceptions import DataframeStreamingException, TimeoutException
from pyarrow.lib import ArrowInvalid
from tests.models.conftest import mock_catalogue_filter_search_response_impl


@pytest.fixture
def instance_file(client):
    client._environment.consumption = 'consumption-test.local'

    instance = client._Instance(
        datafile_id='123',
        total_size=1048576
    )

    client._session.get().json.return_value = {
        'data': [{
            'attributes': {
                'path': 'prefix/test.001.csv',
                'metadata': {}
            }
        }]
    }

    files = instance.files()
    file_ = files[0]

    client.session.reset_mock()

    client._session.get().raw.read.side_effect = [
        b'hello world',
        b''
    ]

    yield file_


class TestDataset:

    @pytest.fixture
    def dataset_load_type_none(self, client):
        return client._DatasetFactory._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'content_type': 'Structured',
                    'location': {},
                    'organisation_short_code': 'abc',
                    'short_code': 'abc',
                    'has_access': True,
                }
            }
        })

    @pytest.fixture
    def dataset_load_type_incremental(self, client):
        return client._DatasetFactory._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'content_type': 'Structured',
                    'location': {},
                    'organisation_short_code': 'abc',
                    'short_code': 'abc',
                    'has_access': True,
                    'load_type': 'Incremental Load',
                }
            }
        })

    @pytest.fixture
    def dataset_load_type_full(self, client):
        return client._DatasetFactory._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'content_type': 'Structured',
                    'location': {},
                    'organisation_short_code': 'abc',
                    'short_code': 'abc',
                    'has_access': True,
                    'load_type': 'Full Load',
                }
            }
        })

    def __setup_mocks(
        self,
        monkeypatch,
        filter_returns_all_prefixes: bool = False,
        has_filter_path_2020: bool = False,
        skip_hidden_files: bool = True,
    ):
        # This has a directory name starting with `.` so must be ignored.
        self.key_latest_0 = MagicMock()
        self.key_latest_0.key = 'abc/.latest/country=fr/' \
                                'nulls_and_unicode.parquet'
        self.key_latest_0.size = 1

        # This has a directory name `as_of=latest` so must be ignored.
        self.key_latest_1 = MagicMock()
        self.key_latest_1.key = 'abc/as_of_date=latest/country=fr/' \
                                'nulls_and_unicode.parquet'
        self.key_latest_1.size = 1

        # This older as_of must be ignored for a full type.
        self.key_2019_0 = MagicMock()
        self.key_2019_0.key = 'abc/as_of_date=2019/country=uk/' \
                              'alltypes_dictionary.parquet'
        self.key_2019_0.size = 1

        self.key_2019_1 = MagicMock()
        self.key_2019_1.key = 'abc/as_of_date=2019/country=us/' \
                              'alltypes_plain.parquet'
        self.key_2019_1.size = 1

        self.key_2020_0 = MagicMock()
        # Note: a key is a relative path starting after the s3://bucket/prefix
        # for this dataset.
        self.key_2020_0.key = 'abc/as_of_date=2020/country=uk/' \
                              'avro.parquet'
        self.key_2020_0.size = 1
        self.key_2020_1 = MagicMock()
        self.key_2020_1.key = 'abc/as_of_date=2020/country=us/' \
                              'nulls_and_unicode.parquet'
        self.key_2020_1.size = 1

        self.filter_mock = MagicMock()

        prefix_2020 = [self.key_2020_0, self.key_2020_1]

        if skip_hidden_files and filter_returns_all_prefixes:
            _filter_side_effects = iter([
                # All files when listing.
                [
                    self.key_2019_0, self.key_2019_1, self.key_2020_0,
                    self.key_2020_1
                ]
            ])
        elif not skip_hidden_files and filter_returns_all_prefixes:
            _filter_side_effects = iter([
                # All files when listing.
                [
                    self.key_latest_0, self.key_latest_1, self.key_2019_0,
                    self.key_2019_1, self.key_2020_0, self.key_2020_1
                ]
            ])
        else:
            _filter_side_effects = iter([
                # All files when listing.
                prefix_2020
            ])

        self.filter_mock.side_effect = _filter_side_effects

        self.objects_mock = MagicMock()
        self.objects_mock.return_value.filter = self.filter_mock

        self.bucket_mock = MagicMock()
        self.bucket_mock.return_value.objects = self.objects_mock()

        self.resource_mock = MagicMock()
        self.resource_mock.return_value.Bucket = self.bucket_mock

        if has_filter_path_2020:
            _common_prefixes = {
                'CommonPrefixes': [
                    {'Prefix': 'abc/as_of_date=2020/'}
                ]
            }
        else:
            _common_prefixes = {
                'CommonPrefixes': [
                    {'Prefix': 'abc/.latest/'},
                    {'Prefix': 'abc/as_of_date=latest/'},
                    {'Prefix': 'abc/as_of_date=2019/'},
                    {'Prefix': 'abc/as_of_date=2020/'},
                ]
            }
        self.resource_mock.return_value.meta.client.list_objects_v2.\
            return_value = _common_prefixes

        self.boto_mock = MagicMock()
        self.boto_mock.return_value.resource.return_value = \
            self.resource_mock()
        monkeypatch.setattr(
            'dli.models.dataset_model.create_refreshing_session',
            self.boto_mock
        )

    def test_dataset_partitions_zero_partitions(
        self,
        monkeypatch,
        dataset_load_type_incremental,
    ):
        """
        Edge case where the S3 structure does not have an partitions.

        :param monkeypatch:
        :return:
        """
        self.__setup_mocks(monkeypatch, filter_returns_all_prefixes=True)

        key_zero_partitions = MagicMock()
        key_zero_partitions.key = 'abc/one/two/alltypes_dictionary.parquet'
        key_zero_partitions.size = 1

        _filter_side_effects = iter([
            [key_zero_partitions]
        ])

        self.filter_mock.side_effect = _filter_side_effects

        partitions = dataset_load_type_incremental.partitions()

        assert partitions == {}

    def test_dataset_partitions_load_type_incremental(
        self,
        dataset_load_type_incremental,
        monkeypatch
    ):
        self.__setup_mocks(monkeypatch, filter_returns_all_prefixes=True)

        partitions = dataset_load_type_incremental.partitions()

        assert partitions == {
            'as_of_date': ['2019', '2020'],
            # Note: we must not get any partitions from the hidden folders.
            'country': ['uk', 'us'],
        }

    def test_dataset_partitions_load_type_full(
        self,
        dataset_load_type_full,
        monkeypatch
    ):
        self.__setup_mocks(monkeypatch, filter_returns_all_prefixes=False)

        partitions = dataset_load_type_full.partitions()

        assert partitions == {
            'as_of_date': ['2020'],
            'country': ['uk', 'us'],
        }

    def test_dataset_instances_is_iterable(self, dataset_load_type_none):
        datafiles = dataset_load_type_none.instances.all()
        assert hasattr(datafiles, '__iter__')

    def test_dataset_download_refers_to_latest_instance(
        self,
        dataset_load_type_incremental,
    ):
        path = './destination_path'

        dataset_load_type_incremental.instances.latest = MagicMock(
            download=MagicMock(return_value='file')
        )

        result = dataset_load_type_incremental.instances.latest.download(path)

        dataset_load_type_incremental.instances.latest.download.\
            assert_called_once_with(path)

        assert result == 'file'

    def test_dataset_instances_make_requests_on_iteration(
        self,
        client,
        dataset_load_type_incremental,
    ):
        client._session.get().json.side_effect = [
            {
                'properties': {'pages_count': 3},
                'entities': [{
                    'properties': {'datasetId': 1}
                }]
            },
            {
                'properties': {'pages_count': 3},
                'entities': [{
                    'properties': {'datasetId': 2}
                }]
            },
            {
                'properties': {'pages_count': 3},
                'entities': [{
                    'properties': {'datasetId': 3}
                }]
            },
        ]

        instances = list(dataset_load_type_incremental.instances.all())
        assert len(instances) == 3

    def test_dataset_instances_make_multiple_iterations(
        self,
        client,
        dataset_load_type_incremental,
    ):
        client._session.get().json.side_effect = [
            {
                'properties': {'pages_count': 3},
                'entities': [{
                    'properties': {'datasetId': 1}
                }]
            },
            {
                'properties': {'pages_count': 3},
                'entities': [{
                    'properties': {'datasetId': 2}
                }]
            },
            {
                'properties': {'pages_count': 3},
                'entities': [{
                    'properties': {'datasetId': 3}
                }]
            },
        ]

        def mock_iter():
            if not hasattr(mock_iter, 'call_count'):
                mock_iter.call_count = 0
                mock_iter.cached = False

            if not mock_iter.cached:
                mock_iter.call_count += 1
                mock_iter.cached = True

            yield from [1, 2, 3]

        instances = list(dataset_load_type_incremental.instances.all())

        assert len(instances) == 3

        instances_from_cache = list(
            dataset_load_type_incremental.instances.all()
        )

        assert len(dataset_load_type_incremental.instances._paginator.cache) \
               == 3
        assert len(instances_from_cache) == 3

    def test_dataset_download_plain(
        self,
        monkeypatch,
        dataset_load_type_incremental,
    ):
        self.__setup_mocks(monkeypatch)

        with tempfile.TemporaryDirectory() as target_dest:
            dataset_load_type_incremental.download(target_dest)

        self.boto_mock.assert_called()
        self.resource_mock.assert_called_once()
        self.bucket_mock.assert_has_calls([
            call('abc')
        ])
        self.objects_mock.assert_called_once()
        self.filter_mock.assert_has_calls([
            # Assert that we only call the non-hidden prefixes.
            call(Prefix='abc/as_of_date'),
        ], any_order=True)

    def test_dataset_download_filter_prefix(
        self,
        monkeypatch,
        dataset_load_type_incremental,
    ):
        self.__setup_mocks(monkeypatch, has_filter_path_2020=True)

        with tempfile.TemporaryDirectory() as target_dest:
            dataset_load_type_incremental.download(
                target_dest,
                filter_path='as_of_date=2020',
            )

        assert (
            self.boto_mock().resource().Bucket()
            .objects.filter.call_args[1]['Prefix']
        ) == 'abc/as_of_date=2020'

    # TODO
    # def test_dataset_is_wrapped_with_analytics_logger(self, client,
    #                                                   monkeypatch,
    #                                                   instance_file):
    #     boto_mock = MagicMock()
    #     monkeypatch.setattr('dli.aws.Session', boto_mock)
    #
    #     dataset = client._DatasetFactory._from_v2_response({
    #         'data': {
    #             'id': '1234',
    #             'attributes': {
    #                 'content_type': 'Structured',
    #                 'location': {},
    #                 'organisation_short_code': 'abc',
    #                 'short_code': 'abc',
    #                 'has_access': True,
    #             }
    #         }
    #     })
    #
    #     with tempfile.TemporaryDirectory() as target_dest:
    #         dataset.download(target_dest)
    #
    #     client._analytics_handler.create_event.call_count = 2
    #     client._analytics_handler.create_event.assert_any_call(
    #         # Not testing the extract user metadata here
    #         ANY, ANY,
    #         'DatasetModel',
    #         '_DatasetModel__download',
    #         {
    #             'self': dataset,
    #             'destination_path': target_dest,
    #             'dataset_id': '1234'
    #         },
    #         result_status_code=200
    #     )

    def test_dataset_info(self, test_client, dataset_request_v2,
                          dictionary_request_v2, fields_request_v2):
        test_client.session.reset_mock()
        test_client._session.get.return_value.json.side_effect = [
            dictionary_request_v2, fields_request_v2
        ]

        dataset = test_client._DatasetFactory._from_v2_response(
            dataset_request_v2
        )

        def redirect(fun):
            with io.StringIO() as buf, redirect_stdout(buf):
                fun()
                output = buf.getvalue()
                return output

        assert(
            redirect(dataset.info)
            .replace(" ", "").replace("\n", "")
            ==
            """
                name    type
                ------  -----------------
                col1    String (Nullable)
                col2    String (Nullable)
                col3    String
            """
            .replace(" ", "").replace("\n", "")
        )

    def test_dataset_list_load_type_none(
        self,
        monkeypatch,
        dataset_load_type_none,
    ):
        self.__setup_mocks(monkeypatch, filter_returns_all_prefixes=True)

        ls = dataset_load_type_none.list()

        self.boto_mock.assert_called_once()
        self.resource_mock.assert_called_once()
        self.bucket_mock.assert_has_calls([
            call('abc')
        ])
        self.objects_mock.assert_called_once()
        self.filter_mock.assert_has_calls([
            # Assert that we only call the non-hidden prefixes.
            call(Prefix='abc/as_of_date'),
        ], any_order=True)

        assert all(type(elem) is str for elem in ls)
        assert ls == [
            # Note: the path is now an absolute path in the format:
            # s3://bucket/prefix/key
            f's3://abc/{self.key_2019_0.key}',
            f's3://abc/{self.key_2019_1.key}',
            f's3://abc/{self.key_2020_0.key}',
            f's3://abc/{self.key_2020_1.key}',
        ], f'Unexpected keys in list: {ls}'

    def test_dataset_list_load_type_incremental(
        self,
        monkeypatch,
        dataset_load_type_incremental,
    ):
        self.__setup_mocks(monkeypatch, filter_returns_all_prefixes=True)

        ls = dataset_load_type_incremental.list()

        self.boto_mock.assert_called_once()
        self.resource_mock.assert_called_once()
        self.bucket_mock.assert_has_calls([
            call('abc')
        ])
        self.objects_mock.assert_called_once()
        self.filter_mock.assert_has_calls([
            # Assert that we only call the non-hidden prefixes.
            call(Prefix='abc/as_of_date'),
        ], any_order=True)

        assert all(type(elem) is str for elem in ls)
        assert ls == [
            # Note: the path is now an absolute path in the format:
            # s3://bucket/prefix/key
            f's3://abc/{self.key_2019_0.key}',
            f's3://abc/{self.key_2019_1.key}',
            f's3://abc/{self.key_2020_0.key}',
            f's3://abc/{self.key_2020_1.key}',
        ], f'Unexpected keys in list: {ls}'

    def test_dataset_list_load_type_full(
        self,
        monkeypatch,
        dataset_load_type_full,
    ):
        self.__setup_mocks(
            monkeypatch,
            # Full Load will only list paths under the most recent `as_of_`
            # common prefix, so it will never see paths from outside that
            # common prefix.
            filter_returns_all_prefixes=False,
        )

        ls = dataset_load_type_full.list()

        self.boto_mock.assert_called_once()
        self.resource_mock.assert_called_once()
        self.bucket_mock.assert_has_calls([
            call('abc')
        ])
        self.objects_mock.assert_called_once()
        self.filter_mock.assert_has_calls([
            # Assert that we only call the most recent common prefix.
            call(Prefix='abc/as_of_date=2020/'),
        ])

        assert all(type(elem) is str for elem in ls)
        assert ls == [
            # Note: the path is now an absolute path in the format:
            # s3://bucket/prefix/key
            f's3://abc/{self.key_2020_0.key}',
            f's3://abc/{self.key_2020_1.key}',
        ], "Expected load type 'Full Load' to have one as_of partition " \
           f"with two files, but list was '{ls}'"

    def test_dataset_list_partition_filter(
        self,
        monkeypatch,
        dataset_load_type_incremental,
    ):
        self.__setup_mocks(monkeypatch, filter_returns_all_prefixes=True)

        ls = dataset_load_type_incremental.list(partitions="as_of_date=2020")

        assert len(ls) == 2

    def test_dataset_list_partition_filter_load_type_full_for_wrong_partition(
        self,
        monkeypatch,
        dataset_load_type_full
    ):
        self.__setup_mocks(monkeypatch, filter_returns_all_prefixes=False)

        ls = dataset_load_type_full.list(partitions="as_of_date=2019")

        assert len(ls) == 0, 'Expected empty list because this as_of_ will' \
                             f'not be inside the full load, but got list: {ls}'

    def test_dataset_list_filter_path(
        self,
        monkeypatch,
        dataset_load_type_incremental,
    ):
        self.__setup_mocks(
            monkeypatch,
            filter_returns_all_prefixes=False,
            has_filter_path_2020=True,
        )

        ls = dataset_load_type_incremental.list(filter_path="as_of_date=2020")

        assert len(ls) == 2

    def test_dataset_list_absolutely(
        self,
        monkeypatch,
        dataset_load_type_incremental,
    ):
        self.__setup_mocks(monkeypatch)

        ls = dataset_load_type_incremental.list(absolute_path=True)

        assert all(map(lambda x: x.startswith("s3://"), ls)), f'List: {ls}'

    def test_dataset_list_relatively(
        self,
        monkeypatch,
        dataset_load_type_incremental,
    ):
        self.__setup_mocks(monkeypatch)

        ls = dataset_load_type_incremental.list(absolute_path=False)

        assert all(map(lambda x: not x.startswith("s3://"), ls)), f'List: {ls}'

    def test_dataset_list_not_skipping_hidden_files(
        self,
        monkeypatch,
        dataset_load_type_incremental,
    ):
        self.__setup_mocks(
            monkeypatch,
            filter_returns_all_prefixes=True,
            skip_hidden_files=False,
        )

        ls = dataset_load_type_incremental.list(skip_hidden_files=False)

        assert len(ls) == 6, 'Expected two from .latest, two from 2020 and ' \
                             f'two from 2020, but got list: {ls}'

    def test_dataset_list_intersecting_partition_and_path(
        self,
        dataset_load_type_incremental,
    ):
        with pytest.raises(ValueError) as excinfo:
            dataset_load_type_incremental.list(
                filter_path="some-filter_path",
                partitions='some-partition',
            )

        message = 'Both `partitions` and `filter_path` are set. Please only ' \
            'provide one of these parameters.'
        assert message in str(excinfo.value)

    def test_dataset_download_intersecting_partition_and_path(
        self,
        dataset_load_type_incremental,
    ):

        with pytest.raises(ValueError) as excinfo:
            dataset_load_type_incremental.download(
                destination_path='.',
                filter_path="some-filter_path",
                partitions=['some-partition'],
            )

        message = 'Both `partitions` and `filter_path` are set. Please only ' \
            'provide one of these parameters.'
        assert message in str(excinfo.value)


class TestFileDownload:

    def test_instance_file_download_simple(self, instance_file):
        old_dir = os.path.abspath(os.curdir)

        with tempfile.TemporaryDirectory() as target_dest:
            os.chdir(os.path.abspath(target_dest))
            instance_file.download()
            with open(os.path.join(target_dest, 'prefix/test.001.csv')) as f:
                assert f.read() == 'hello world'

        os.chdir(old_dir)

    def test_instance_file_download_with_path(self, instance_file):
        with tempfile.TemporaryDirectory() as target_dest:
            instance_file.download(target_dest)
            with open(os.path.join(target_dest, 'prefix/test.001.csv')) as f:
                assert f.read() == 'hello world'

    def test_instance_file_download_non_existent_path(
        self, instance_file
    ):
        with tempfile.TemporaryDirectory() as target_dest:
            target_dest = os.path.join(target_dest, 'a/b/c/')
            instance_file.download(target_dest)
            with open(os.path.join(
                target_dest, 'prefix/test.001.csv'
            )) as f:
                assert f.read() == 'hello world'

    def test_instance_file_download_non_existent_file(
        self, instance_file
    ):
        with tempfile.TemporaryDirectory() as target_dest:
            target_dest = os.path.join(target_dest, 'a')
            instance_file.download(target_dest)
            with open(os.path.join(
                target_dest, 'prefix/test.001.csv'
            )) as f:
                assert f.read() == 'hello world'

    def test_instance_file_overwrites(
        self, client, instance_file
    ):
        with tempfile.TemporaryDirectory() as target_dest:
            target_dest = os.path.join(target_dest, 'a')
            instance_file.download(target_dest)
            with open(os.path.join(
                target_dest, 'prefix/test.001.csv'
            )) as f:
                assert f.read() == 'hello world'

        client.session.reset_mock()
        client._session.get().raw.read.side_effect = [
            b'brave new world',
            b''
        ]

        with tempfile.TemporaryDirectory() as target_dest:
            target_dest = os.path.join(target_dest, 'a')
            instance_file.download(target_dest)
            with open(os.path.join(
                target_dest, 'prefix/test.001.csv'
            )) as f:
                assert f.read() == 'brave new world'


class TestInstanceDownload:

    @pytest.fixture
    def instance(self, client):
        client._environment.consumption = 'consumption-test.local'

        instance = client._Instance(
            datafile_id='123'
        )

        client._session.get().json.return_value = {
            'data': [{
                'attributes': {
                    'path': f's3://bucket/prefix/test.00{i}.csv',
                    'metadata': {}
                }
            } for i in range(3)
            ]
        }

        yield instance

    def test_instance_download_all_should_refer_to_download_function(
        self,
        instance
    ):
        path = './target'
        instance.download = MagicMock(return_value='file')
        result = instance.download_all(path, flatten=False)

        instance.download.assert_called_once_with(path, False)
        assert result == 'file'

    def test_instance_download_flatten(self, client, instance):
        client._session.get().raw = io.BytesIO(b'abc')

        with tempfile.TemporaryDirectory() as target_dest:
            to = os.path.join(target_dest, 'a')
            result = instance.download_all(to, flatten=True)
            directory_of_flattened = result[0].split('/test')[0]
            list_dir = sorted(os.listdir(directory_of_flattened))

            assert (
                list_dir ==
                ['test.000.csv', 'test.001.csv', 'test.002.csv']
            )


class TestInstanceDataframe:

    @pytest.fixture
    def arrow_ipc_stream(self) -> (pandas.DataFrame, io.BytesIO):
        # Rows:
        # 100 -> 205.5 kB (205,548 bytes) on disk in 0.44 seconds.
        # 10000 -> 9.2 MB (9,234,131 bytes) on disk in 0.67 seconds.

        max_rows = 10000
        # Max columns should match a worst case. For Transaction Analysis
        # there are over 369 columns.
        max_columns = 372

        num_column_types = 6

        # Seed the random number generator so that we can reproduce the output.
        numpy.random.seed(0)

        def bools_df():
            return pandas.DataFrame(
                numpy.random.choice(
                    [False, True],
                    size=(max_rows, int(max_columns / num_column_types))
                ),
                columns=[str(f'column_bool_{x}') for x in
                         range(int(max_columns / num_column_types))],
                dtype='bool',
            )

        def ints_df():
            return pandas.DataFrame(
                numpy.random.randint(
                    low=0,
                    high=1000000,
                    size=(max_rows, int(max_columns / num_column_types))
                ),
                columns=[str(f'column_int_{x}') for x in
                         range(int(max_columns / num_column_types))],
            )

        def floats_df():
            return pandas.DataFrame(
                numpy.random.rand(
                    max_rows,
                    int(max_columns / num_column_types)),
                columns=[str(f'column_float_{x}') for x in
                         range(int(max_columns / num_column_types))],
            )

        def string_df():
            import pandas.util.testing
            return pandas.DataFrame(
                numpy.random.choice(
                    pandas.util.testing.RANDS_CHARS,
                    size=(max_rows, int(max_columns / num_column_types))
                ),
                columns=[str(f'column_str_{x}') for x in
                         range(int(max_columns / num_column_types))],
            )

        def dates_df():
            dates = pandas.date_range(start='1970-01-01', end='2020-01-01')

            return pandas.DataFrame(
                numpy.random.choice(
                    dates,
                    size=(max_rows, int(max_columns / num_column_types))
                ),
                columns=[str(f'column_date_{x}') for x in
                         range(int(max_columns / num_column_types))],
            )

        def datetimes_df():
            datetimes = pandas.date_range(start='1970-01-01', end='2020-01-01',
                                          freq='H')

            return pandas.DataFrame(
                numpy.random.choice(
                    datetimes,
                    size=(max_rows, int(max_columns / num_column_types))
                ),
                columns=[str(f'column_datetime_{x}') for x in
                         range(int(max_columns / num_column_types))],
            )

        df = pandas.concat(
            [
                bools_df(),
                datetimes_df(),
                ints_df(), floats_df(), string_df(), dates_df(),
            ],
            # Concat columns axis
            axis=1,
            copy=False
        )

        # Randomise the column order so the compression cannot optimise
        # easily based on types as this will mix float columns among the int
        # columns.
        random_column_order = df.columns.tolist()
        random.shuffle(random_column_order)
        df = df[random_column_order]

        assert df.shape[0] == max_rows
        assert df.shape[1] == max_columns

        sink = pyarrow.BufferOutputStream()
        batch = pyarrow.RecordBatch.from_pandas(df=df, nthreads=1)
        writer = pyarrow.RecordBatchStreamWriter(sink, batch.schema)
        writer.write_batch(batch)
        # It's important the IPC stream is closed
        writer.close()
        var = sink.getvalue()
        return df, io.BytesIO(var)

    @pytest.fixture
    def dataset(self, client):
        return client._DatasetFactory._from_v2_response({
            'data': {
                'id': '1234',
                'attributes': {
                    'content_type': 'Structured',
                    'location': {},
                    'has_access': True,
                    'short_code': 'some-short-code',
                }
            }
        })

    def test_dataset_dataframe(self, benchmark, arrow_ipc_stream,
                               dataset):

        test_dataframe, stream = arrow_ipc_stream
        dataset._client = MagicMock()
        dataset._client._environment.consumption = 'https://test/'
        dataset._client.session.get().status_code = 200
        msg = json.dumps({
            'status': 200,
        }).encode('utf')

        _io_bytesio_seekend = 2
        stream.seek(0, _io_bytesio_seekend)
        stream.write(msg)

        def _bench():
            stream.seek(0)
            # reset our fake socket
            dataset._client.session.get().raw = stream
            return dataset.dataframe()

        # The benchmark code will run several times to get an average runtime.
        df = benchmark(_bench)

        # Check that we can read from the file that was written.
        pandas.testing.assert_frame_equal(test_dataframe, df)

    def test_dataframe_handle_errors(self, arrow_ipc_stream,
                                     dataset):
        test_dataframe, stream = arrow_ipc_stream
        error_msg = json.dumps({
            'status': 400,
        }).encode('utf')

        _io_bytesio_seekend = 2
        stream.seek(0, _io_bytesio_seekend)
        stream.write(error_msg)
        stream.seek(0)

        dataset._client = MagicMock()
        dataset._client._environment.consumption = 'https://test/'

        dataset._client.session.get().raw = stream
        dataset._client.session.get().status_code = 200

        with pytest.raises(DataframeStreamingException):
            dataset.dataframe(1)

    def test_dataframe_handle_arrowinvalid_exception(
            self, arrow_ipc_stream, dataset
    ):
        arrow_ipc_stream = io.BytesIO(b'invalid')

        arrow_ipc_stream.seek(0)

        dataset._client = MagicMock()
        dataset._client._environment.consumption = 'https://test/'

        dataset._client.session.get().raw = arrow_ipc_stream
        dataset._client.session.get().status_code = 200

        with pytest.raises(ArrowInvalid) as excinfo:
            dataset.dataframe(1)

        message = r'Sorry, the dataframe you are trying to read is too ' \
                  'large to fit into memory. Please consider one of these ' \
                  'alternatives:' \
                  '\n\n1. Run the same code on a machine that has more ' \
                  'RAM, for example a Jupyter Notebook hosted on an AWS ' \
                  'environment.'\
                  '\n2. Use a big data tool such as Spark, Dask or similar ' \
                  'to read the data via our S3 proxy. Unlike Pandas, these ' \
                  'big data tools are able to apply some operations such ' \
                  'as filtering without having to hold all of the data ' \
                  'into memory. See https://supportlibrary.ihsmarkit.com/' \
                  'display/DLSL/Using+the+S3+Proxy+with+Big+Data+tooling' \
                  '\n\nOriginal exception: '

        error_message = str(excinfo.value)
        assert error_message.startswith(message)

        pattern = r"Original exception: Expected to read \d+ metadata " \
                  r"bytes, but only read \d+"
        assert re.search(pattern, error_message)


class TestDatasetModule:

    def test_dataset_contents(self, capsys, test_client,
                              package_request_index_v2,
                              dataset_request_index_v2,
                              instance_request_v1):
        test_client._session.get.return_value.json.side_effect = [
            package_request_index_v2,
            dataset_request_index_v2, instance_request_v1,
        ]

        pp = test_client.packages(search_term='', only_mine=False)
        ds = pp['Test Package'].datasets()
        d = ds['TestDataset']
        d.contents()
        captured = capsys.readouterr()
        assert captured.out == '\nINSTANCE 1 (size: 1.00 Mb)\n'

    @pytest.fixture()
    def return_value(self, dataset_request_index_v2):
        return partial(
            mock_catalogue_filter_search_response_impl,
            dataset_request_index_v2
        )

    def test_find_no_params(self, test_client, return_value):
        test_client._session.get = return_value
        assert len(test_client.datasets()) == 2

    def test_find_datasets_no_list(self, test_client, return_value):
        test_client._session.get = return_value
        assert len(test_client.datasets(search_term=[])) == 2

    def test_find_datasets_all_data(self, test_client, return_value):
        test_client._session.get = return_value
        assert len(test_client.datasets(only_mine=False)) == 2

    def test_find_datasets_only_mine(self, test_client, return_value):
        test_client._session.get = return_value
        assert len(test_client.datasets(only_mine=True)) == 1

    def test_find_datasets_no_search_all_data(self, test_client, return_value):
        test_client._session.get = return_value
        assert len(test_client.datasets(search_term=None, only_mine=False)) \
               == 2

    def test_find_datasets_no_list_all_data(self, test_client, return_value):
        test_client._session.get = return_value
        assert len(test_client.datasets(search_term=[], only_mine=False)) == 2

    def test_find_datasets_no_term_only_mine(self, test_client, return_value):
        test_client._session.get = return_value
        assert len(test_client.datasets(search_term=None, only_mine=True)) == 1

    def test_find_datasets_no_list_only_mine(self, test_client, return_value):
        test_client._session.get = return_value
        assert len(test_client.datasets(search_term=[], only_mine=True)) == 1

    def test_find_datasets_ignore_bad_data_list(
        self,
        test_client,
        return_value
    ):
        test_client._session.get = return_value
        assert len(test_client.datasets(
            search_term=["baddata"], only_mine=False)) == 2

    def test_find_datasets_ignore_multi_bad_data(
        self,
        test_client,
        return_value
    ):
        test_client._session.get = return_value
        assert len(test_client.datasets(
            search_term=["baddata", "baddata"], only_mine=False)) == 2

    def test_find_datasets_ignore_bad_keep_good_data(
        self,
        test_client,
        return_value
    ):
        test_client._session.get = return_value
        assert len(
            test_client.datasets(
                search_term=["short_code=TestDataset", "baddata"],
                only_mine=False
            )
        ) == 1

    def test_find_datasets_ignore_bad_data(self, test_client, return_value):
        test_client._session.get = return_value
        assert len(test_client.datasets(
            search_term="baddata", only_mine=False)) == 2

    def test_find_datasets_name_search_match(self, test_client, return_value):
        test_client._session.get = return_value
        # testing search term with only_mine param against valid search term
        assert len(test_client.datasets(
            search_term=["short_code=TestDataset"], only_mine=False)) == 1

    def test_find_datasets_no_search_my_data(self, test_client, return_value):
        test_client._session.get = return_value
        assert len(test_client.datasets(
            search_term=[], only_mine=True)) == 1

    def test_find_datasets_no_match(self, test_client, return_value):
        test_client._session.get = return_value
        assert len(test_client.datasets(
            search_term=["short_code=NotTestDataset"], only_mine=True)) == 0

    def test_find_datasets_multi_param_my_data(
        self,
        test_client,
        return_value
    ):
        test_client._session.get = return_value
        # multi-param against only_mine
        assert len(test_client.datasets(
            search_term=["name=Other Dataset"],
            only_mine=True)) == 0

    def test_find_datasets_multi_param_all_data(
        self,
        test_client,
        return_value
    ):
        test_client._session.get = return_value
        assert len(test_client.datasets(
            search_term=["name=Test DataSet", "description=description",
                         "organisation_name=IHS Markit"],
            only_mine=False)) == 1

    def test_find_datasets_find_both(self, test_client, return_value):
        test_client._session.get = return_value
        assert len(test_client.datasets(
            search_term=["description=description"],
            only_mine=False)) == 2

    def test_find_datasets_multi_param_no_match(
        self,
        test_client,
        return_value
    ):
        test_client._session.get = return_value
        # multi-param with bad data
        assert len(test_client.datasets(
            search_term=[
                "description=description",
                "organisation_name=baddata"
            ],
            only_mine=False)) == 0

    def test_find_datasets_like_name(self, test_client, return_value):
        test_client._session.get = return_value
        # 'like' test
        assert len(test_client.datasets(
            search_term=["name like Test"],
            only_mine=True)) == 1

    def test_find_datasets_by_id(self, test_client, return_value):
        test_client._session.get = return_value
        # 'like' test
        assert len(test_client.datasets(
            search_term=["id=5b01376e-975d-11e9-8832-7e5ef76d533f"],
            only_mine=True)) == 1

    def test_find_datasets_override_has_access_return_none(
        self,
        test_client,
        return_value
    ):
        test_client._session.get = return_value
        # test override of param against kwarg of same (accomplish both ways)
        assert len(test_client.datasets(
            search_term=["has_access=True"],
            only_mine=False)) == 0

    def test_datasets_called_twice(self, test_client, package_request_index_v2,
                                   dataset_request_index_v2):
        test_client._session.get.return_value.json.side_effect = [
            package_request_index_v2, dataset_request_index_v2
        ]
        pp = test_client.packages(search_term='', only_mine=False)

        # This should work twice
        assert pp['Test Package'].datasets()
        assert pp['Test Package'].datasets()

    def test_retrieve_instance_given_package_dataset(self, test_client,
                                                     package_request_index_v2,
                                                     dataset_request_index_v2,
                                                     instance_request_v1
                                                     ):
        test_client._session.get.return_value.json.side_effect = [
            package_request_index_v2,
            dataset_request_index_v2, instance_request_v1,
        ]

        pp = test_client.packages(search_term='', only_mine=False)
        ds = pp['Test Package'].datasets()
        d = ds['TestDataset']
        instances = list(d.instances.all())
        assert (len(instances) == 1)
        assert (instances[0].datafile_id == 1)

    def test_retrieve_dataset_by_get(
        self,
        test_client,
        visible_orgs_v2,
        unstructured_dataset_request_index_v2
    ):
        test_client._session.get.return_value.json.side_effect = [
            visible_orgs_v2,
            unstructured_dataset_request_index_v2,
        ]

        test_ds = test_client.datasets.get('TestDataset')
        assert test_ds.short_code == "TestDataset"



#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import logging
import sys
import os

from logging.handlers import RotatingFileHandler
from pathlib import Path

from pythonjsonlogger import jsonlogger  # noqa: I900


def _setup_logging(where_level):
    logger = logging.getLogger(__name__)

    # Simple print logger
    tracing = os.environ.get('DLI_TRACE')
    if tracing and tracing.lower() == 'true':
        trace_logger = logging.getLogger('trace_logger')
        trace_logger.setLevel(logging.DEBUG)

        if not trace_logger.hasHandlers():
            s_handler = logging.StreamHandler(stream=sys.stderr)
            s_handler.setLevel(logging.DEBUG)
            trace_logger.addHandler(s_handler)

    json_format = jsonlogger.JsonFormatter(
        '%(asctime)s - %(levelname)s - %(name)s - %(message)s',
        json_indent=2
    )

    switcher = {
        "error": logging.ERROR,
        "warn": logging.WARNING,
        "debug": logging.DEBUG,
        "info": logging.INFO
    }

    to, _, level = str(where_level).partition(":")
    chosen_level = switcher.get(level.lower(), logging.DEBUG)
    logger.setLevel(chosen_level)

    s_handler_formatter = logging.Formatter(
        '%(asctime)s - %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )

    if to == "stdout":
        s_handler = logging.StreamHandler(stream=sys.stdout)
        s_handler.setLevel(chosen_level)
        s_handler.setFormatter(s_handler_formatter)
        logger.addHandler(s_handler)
    if to == "stderr":
        s_handler = logging.StreamHandler(stream=sys.stderr)
        s_handler.setLevel(chosen_level)
        s_handler.setFormatter(s_handler_formatter)

        if not logger.hasHandlers():
            logger.addHandler(s_handler)
    else:
        logging.StreamHandler(stream=None)



    log_folder = "logs/"
    Path(log_folder).mkdir(parents=True, exist_ok=True)

    r_handler = RotatingFileHandler(
        f'{log_folder}sdk.log', mode='w', backupCount=3)
    r_handler.setLevel(logging.DEBUG)
    r_handler.setFormatter(json_format)

    if len(logger.handlers) == 1:
        logger.addHandler(r_handler)

    return logger



from unittest.mock import MagicMock, patch
from dli.client import session as sm


class TestVersionCheck:

    def test_ssl_error_in_version_check_caught(self):
        # going to same server as pypi but not with a valid hostname.
        with patch('dli.client.session.trace_logger') as logger_mock:
            sm._set_pypi_url = MagicMock(
                return_value='https://151.101.60.223/pypi/dli/json'
            )
            sm.version_check(package_name='dli')
            logger_mock.warning.assert_called()



#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import os
from distutils.version import StrictVersion
import warnings
import requests
import logging

from dli.client import _setup_logging
from dli.client.dli_client import DliClient
from dli import __version__


logger = logging.getLogger(__name__)
trace_logger = logging.getLogger('trace_logger')
DEPRECATED_AGE = 1


def _set_pypi_url(package_name):
    return f"https://pypi.org/pypi/{package_name}/json"


def _is_release(x):
    return x.replace(".", "").isnumeric()


def version_check(package_name):
    url = _set_pypi_url(package_name)

    try:
        data = requests.get(url).json()
        versions = list(x for x in data["releases"].keys() if _is_release(x))
        versions.sort(key=StrictVersion)
        versions.reverse()

        offset = versions.index(__version__)
        if offset > DEPRECATED_AGE:
            warnings.warn("You are using an old version of the SDK, please "
                          "upgrade using `pip install dli --upgrade` "
                          "before the SDK no longer functions as expected",
                          PendingDeprecationWarning)

    except requests.exceptions.SSLError as e:
        # Version check has been see to fail by Gilberto with an SSLError
        # when doing the handshake with pypi.
        trace_logger.warning(
            f'Could not connect to pypi to check the installed DLI '
            f'version. This may be due to your internet '
            f'security rules. You can check for the latest version by '
            f'visiting https://pypi.org/project/dli/',
            extra={'SSLError': e}
        )
        pass
    except Exception:
        pass


def _start_session(
    api_key=None,
    root_url="https://catalogue.datalake.ihsmarkit.com/__api",
    host=None,
    debug=False,
    strict=True,
    use_keyring=True,
    log_level=None,
):
    if log_level is None:
        log_level = 'stderr:info'
    logger = _setup_logging(log_level)

    version_check('dli')

    user = os.environ.get("DLI_ACCESS_KEY_ID")
    pasw = os.environ.get("DLI_SECRET_ACCESS_KEY")
    api_key = api_key or os.environ.get("DLI_API_KEY")

    logger.debug("Checking auth flow")

    try:
        if (user is not None and pasw is not None) and api_key is None:
            logger.debug("Using SAM client credentials authentication flow")
            return get_client()(root_url, host, debug=debug, strict=strict,
                                access_id=user, secret_key=pasw, use_keyring=use_keyring)
        elif (user is None or pasw is None) and api_key is None:
            logger.debug("Using SAM PKCE web authentication flow")
            return get_client()(root_url, host, debug=debug, strict=strict, use_keyring=use_keyring)
        elif api_key is not None:
            logger.debug("Using API Key authentication flow")
            if api_key is not None:
                return get_client()(root_url, host, debug=debug, strict=strict,
                                    secret_key=api_key, use_keyring=use_keyring)
    except SystemExit:
        return


def start_session(*args, **kwargs):
    """
    .. deprecated:: 1.8.3

        1. `api_key`: please use the basic flow or client credentials.

        2. Please switch to using `client.connect()` as we may remove this
        in the future.


    Entry point for the Data Lake SDK, returns a client instance that
    can be used to consume or register datasets.

    Example for creating a client:

        from dli.client import session
        client = session.start_session()

    :param str api_key: Your API key, can be retrieved from your dashboard in
                        the Catalogue UI.
    :param str root_url: Optional. The environment you want to point to. By default it
                        points to Production.
    :param str host: Optional. Advanced usage, meant to force a hostname when DNS resolution
                     is not reacheable from the user's network.
                     This is meant to be used in conjunction with an
                     IP address as the root url.
                     Example: catalogue.datalake.ihsmarkit.com

    :param bool debug: Optional. Log SDK operations to a file in the current working
                       directory with the format "sdk-{end of api key}-{timestamp}.log"

    :param bool strict: Optional. When True, all exception messages and stack
                        trace are printed. When False, a shorter message is
                        printed and `None` should be returned.
    :param bool use_keyring: Optional. When True, cache the JWT in the system keyring
                             and retrieve that JWT from the system keyring if set.
                             Otherwise disable the keyring completely.

    :returns: Data Lake interface client
    :rtype: dli.client.dli_client.DliClient

    .. deprecated:: 1.8.3

        `api_key`: please use the basic flow or client credentials.

    """

    # We know that headless and Excel spreadsheets still use this call, and
    # the users do not look at the deprecation message so
    # we cannot remove. We are going to not reference this in out
    # documentation so that new users do not use it. When we have all our
    # new users using `connect` we can remove this.

    # We know that headless and Excel spreadsheets still use this call, and
    # the users do not look at the deprecation message so
    # we cannot remove. We are going to not reference this in out
    # documentation so that new users do not use it. When we have all our
    # new users using `connect` we can remove this.

    # We know that headless and Excel spreadsheets still use this call, and
    # the users do not look at the deprecation message so
    # we cannot remove. We are going to not reference this in out
    # documentation so that new users do not use it. When we have all our
    # new users using `connect` we can remove this.

    return _start_session(*args, **kwargs)


def get_client():
    return DliClient
